"""Tests for pz_async_vec_env.py and pz_vec_env.py"""

import multiprocessing as mp
import os
import signal
import time
from multiprocessing import Process
from multiprocessing.sharedctypes import SynchronizedArray
from unittest.mock import patch

import gymnasium as gym
import gymnasium.spaces as spaces
import numpy as np
import pytest
from gymnasium.error import (
    AlreadyPendingCallError,
    ClosedEnvironmentError,
    NoAsyncCallError,
)
from gymnasium.spaces import Box, Discrete, MultiDiscrete
from gymnasium.vector.utils import CloudpickleWrapper
from pettingzoo import ParallelEnv
from pettingzoo.mpe import simple_speaker_listener_v4
from pettingzoo.sisl import pursuit_v4

from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer
from agilerl.vector.pz_async_vec_env import (  # PettingZooExperienceSpec,; SharedMemory,
    AsyncPettingZooVecEnv,
    AsyncState,
    Observations,
    _async_worker,
    get_placeholder_value,
    write_to_shared_memory,
)
from agilerl.vector.pz_vec_env import PettingZooVecEnv
from tests.pz_vector_test_utils import GenericTestEnv, term_env


class DummyRecv:
    def __init__(self, cmd, data):
        self.call_count = 0
        self.cmd = cmd
        self.data = data

    def __call__(self, *args, **kwargs):
        self.call_count += 1
        if self.call_count > 1:
            return "close", None
        else:
            return self.cmd, self.data


class DictSpaceTestEnv(ParallelEnv):
    """Test environment with dictionary observation spaces"""

    metadata = {"render_modes": ["human", "rgb_array"], "name": "dict_space_test_v0"}

    def __init__(self, render_mode=None):
        self.possible_agents = ["agent_0", "other_agent_0"]
        self.agents = self.possible_agents.copy()
        self.render_mode = render_mode

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents.copy()
        observations = {
            "agent_0": {
                "position": np.array([0.1, 0.2, 0.3], dtype=np.float32),
                "velocity": np.array([0.01, 0.02], dtype=np.float32),
            },
            "other_agent_0": {
                "position": np.array([0.4, 0.5, 0.6], dtype=np.float32),
                "velocity": np.array([0.03, 0.04], dtype=np.float32),
            },
        }
        infos = {agent: {} for agent in self.agents}
        return observations, infos

    def step(self, actions):
        observations = {
            "agent_0": {
                "position": np.array([0.2, 0.3, 0.4], dtype=np.float32),
                "velocity": np.array([0.02, 0.03], dtype=np.float32),
            },
            "other_agent_0": {
                "position": np.array([0.5, 0.6, 0.7], dtype=np.float32),
                "velocity": np.array([0.04, 0.05], dtype=np.float32),
            },
        }
        rewards = {agent: 1.0 for agent in self.agents}
        terminations = {agent: False for agent in self.agents}
        truncations = {agent: False for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        return observations, rewards, terminations, truncations, infos

    def observation_space(self, agent):
        return spaces.Dict(
            {
                "position": Box(low=-1.0, high=1.0, shape=(3,), dtype=np.float32),
                "velocity": Box(low=-0.1, high=0.1, shape=(2,), dtype=np.float32),
            }
        )

    def action_space(self, agent):
        return Box(low=-1.0, high=1.0, shape=(2,), dtype=np.float32)

    def render(self):
        if self.render_mode == "rgb_array":
            return np.ones((64, 64, 3), dtype=np.uint8)
        return None

    def close(self):
        pass


class TupleSpaceTestEnv(ParallelEnv):
    """Test environment with tuple observation spaces"""

    metadata = {"render_modes": ["human", "rgb_array"], "name": "tuple_space_test_v0"}

    def __init__(self, render_mode=None):
        self.possible_agents = ["agent_0", "other_agent_0"]
        self.agents = self.possible_agents.copy()
        self.render_mode = render_mode

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents.copy()
        observations = {
            "agent_0": (
                np.array([0.1, 0.2, 0.3], dtype=np.float32),
                np.array([0.01, 0.02], dtype=np.float32),
            ),
            "other_agent_0": (
                np.array([0.4, 0.5, 0.6], dtype=np.float32),
                np.array([0.03, 0.04], dtype=np.float32),
            ),
        }
        infos = {agent: {} for agent in self.agents}
        return observations, infos

    def step(self, actions):
        observations = {
            "agent_0": (
                np.array([0.2, 0.3, 0.4], dtype=np.float32),
                np.array([0.02, 0.03], dtype=np.float32),
            ),
            "other_agent_0": (
                np.array([0.5, 0.6, 0.7], dtype=np.float32),
                np.array([0.04, 0.05], dtype=np.float32),
            ),
        }
        rewards = {agent: 1.0 for agent in self.agents}
        terminations = {agent: False for agent in self.agents}
        truncations = {agent: False for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        return observations, rewards, terminations, truncations, infos

    def observation_space(self, agent):
        return spaces.Tuple(
            (
                Box(low=-1.0, high=1.0, shape=(3,), dtype=np.float32),
                Box(low=-0.1, high=0.1, shape=(2,), dtype=np.float32),
            )
        )

    def action_space(self, agent):
        return Box(low=-1.0, high=1.0, shape=(2,), dtype=np.float32)

    def render(self):
        if self.render_mode == "rgb_array":
            return np.ones((64, 64, 3), dtype=np.uint8)
        return None

    def close(self):
        pass


class ComplexDictSpaceTestEnv(ParallelEnv):
    """Test environment with dictionary observation spaces containing both vector and image data"""

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "name": "complex_dict_space_test_v0",
    }

    def __init__(self, render_mode=None):
        self.possible_agents = ["agent_0", "other_agent_0"]
        self.agents = self.possible_agents.copy()
        self.render_mode = render_mode

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents.copy()
        observations = {
            "agent_0": {
                "position": np.array([0.1, 0.2, 0.3], dtype=np.float32),
                "velocity": np.array([0.01, 0.02], dtype=np.float32),
                "image": np.ones((16, 16, 3), dtype=np.uint8) * 100,
            },
            "other_agent_0": {
                "position": np.array([0.4, 0.5, 0.6], dtype=np.float32),
                "velocity": np.array([0.03, 0.04], dtype=np.float32),
                "image": np.ones((16, 16, 3), dtype=np.uint8) * 200,
            },
        }
        infos = {agent: {} for agent in self.agents}
        return observations, infos

    def step(self, actions):
        observations = {
            "agent_0": {
                "position": np.array([0.2, 0.3, 0.4], dtype=np.float32),
                "velocity": np.array([0.02, 0.03], dtype=np.float32),
                "image": np.ones((16, 16, 3), dtype=np.uint8) * 150,
            },
            "other_agent_0": {
                "position": np.array([0.5, 0.6, 0.7], dtype=np.float32),
                "velocity": np.array([0.04, 0.05], dtype=np.float32),
                "image": np.ones((16, 16, 3), dtype=np.uint8) * 250,
            },
        }
        rewards = {agent: 1.0 for agent in self.agents}
        terminations = {agent: False for agent in self.agents}
        truncations = {agent: False for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        return observations, rewards, terminations, truncations, infos

    def observation_space(self, agent):
        return spaces.Dict(
            {
                "position": Box(low=-1.0, high=1.0, shape=(3,), dtype=np.float32),
                "velocity": Box(low=-0.1, high=0.1, shape=(2,), dtype=np.float32),
                "image": Box(low=0, high=255, shape=(16, 16, 3), dtype=np.uint8),
            }
        )

    def action_space(self, agent):
        return Box(low=-1.0, high=1.0, shape=(2,), dtype=np.float32)

    def render(self):
        if self.render_mode == "rgb_array":
            return np.ones((64, 64, 3), dtype=np.uint8)
        return None

    def close(self):
        pass


class ComplexTupleSpaceTestEnv(ParallelEnv):
    """Test environment with tuple observation spaces containing both vector and image data"""

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "name": "complex_tuple_space_test_v0",
    }

    def __init__(self, render_mode=None):
        self.possible_agents = ["agent_0", "other_agent_0"]
        self.agents = self.possible_agents.copy()
        self.render_mode = render_mode

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents.copy()
        observations = {
            "agent_0": (
                np.array([0.1, 0.2, 0.3], dtype=np.float32),
                np.array([0.01, 0.02], dtype=np.float32),
                np.ones((16, 16, 3), dtype=np.uint8) * 100,
            ),
            "other_agent_0": (
                np.array([0.4, 0.5, 0.6], dtype=np.float32),
                np.array([0.03, 0.04], dtype=np.float32),
                np.ones((16, 16, 3), dtype=np.uint8) * 200,
            ),
        }
        infos = {agent: {} for agent in self.agents}
        return observations, infos

    def step(self, actions):
        observations = {
            "agent_0": (
                np.array([0.2, 0.3, 0.4], dtype=np.float32),
                np.array([0.02, 0.03], dtype=np.float32),
                np.ones((16, 16, 3), dtype=np.uint8) * 150,
            ),
            "other_agent_0": (
                np.array([0.5, 0.6, 0.7], dtype=np.float32),
                np.array([0.04, 0.05], dtype=np.float32),
                np.ones((16, 16, 3), dtype=np.uint8) * 250,
            ),
        }
        rewards = {agent: 1.0 for agent in self.agents}
        terminations = {agent: False for agent in self.agents}
        truncations = {agent: False for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        return observations, rewards, terminations, truncations, infos

    def observation_space(self, agent):
        return spaces.Tuple(
            (
                Box(low=-1.0, high=1.0, shape=(3,), dtype=np.float32),
                Box(low=-0.1, high=0.1, shape=(2,), dtype=np.float32),
                Box(low=0, high=255, shape=(16, 16, 3), dtype=np.uint8),
            )
        )

    def action_space(self, agent):
        return Box(low=-1.0, high=1.0, shape=(2,), dtype=np.float32)

    def render(self):
        if self.render_mode == "rgb_array":
            return np.ones((64, 64, 3), dtype=np.uint8)
        return None

    def close(self):
        pass


def actions_to_list_helper(actions):
    passed_actions_list = [[] for _ in list(actions.values())[0]]
    for env_idx, _ in enumerate(list(actions.values())[0]):
        for possible_agent in actions.keys():
            passed_actions_list[env_idx].append(actions[possible_agent][env_idx])

    return passed_actions_list


# @pytest.fixture
# def pz_experience_spec():
#     return PettingZooExperienceSpec(8)


@pytest.fixture(autouse=True)
def clean_process_fixture():
    """Fixture to ensure processes are cleaned up between tests"""
    # Before each test
    yield
    # After each test - forcibly terminate any stray processes
    for p in mp.active_children():
        p.terminate()
        p.join(timeout=1.0)


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(8)]]
)
def test_create_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    assert env.single_action_space
    assert env.action_space
    assert env.single_observation_space
    assert env.observation_space
    assert env.num_envs == 8
    for val in env._obs_buffer.values():
        assert isinstance(val, SynchronizedArray)
    assert isinstance(env.observations, Observations)
    assert env.processes
    env.reset()
    env.close()


@pytest.mark.parametrize("seed", [1, None])
@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(8)]]
)
def test_reset_async_pz_vector_env(seed, env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    agents = env.possible_agents[:]
    observations, infos = env.reset(seed=seed)
    env.close()
    for agent in agents:
        assert isinstance(env.observation_space(agent), Box)
        assert isinstance(observations[agent], np.ndarray)
        assert observations[agent].dtype == env.observation_space(agent).dtype
        assert (
            observations[agent].shape
            == (8,) + env.single_observation_space(agent).shape
        )
        assert observations[agent].shape == env.observation_space(agent).shape
    assert isinstance(infos, dict)
    assert set(agents).issubset(set(infos.keys()))

    try:
        env_fns = [simple_speaker_listener_v4.parallel_env for _ in range(8)]
        env = AsyncPettingZooVecEnv(env_fns)
    finally:
        env.close()
    for agent in agents:
        assert isinstance(env.observation_space(agent), Box)
        assert isinstance(observations[agent], np.ndarray)
        assert observations[agent].dtype == env.observation_space(agent).dtype
        assert (
            observations[agent].shape
            == (8,) + env.single_observation_space(agent).shape
        )
        assert observations[agent].shape == env.observation_space(agent).shape
        assert set(agents).issubset(set(infos.keys()))


@pytest.mark.parametrize(
    "env_fns",
    [
        [
            lambda: simple_speaker_listener_v4.parallel_env(render_mode="rgb_array")
            for _ in range(8)
        ]
    ],
)
def test_render_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    assert env.render_mode == "rgb_array"

    env.reset()
    rendered_frames = env.render()
    assert isinstance(rendered_frames, tuple)
    assert len(rendered_frames) == env.num_envs
    assert all(isinstance(frame, np.ndarray) for frame in rendered_frames)
    env.close()


@pytest.mark.parametrize("use_single_action_space", [False, True])
@pytest.mark.parametrize(
    "env_fns",
    [
        [
            lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
            for _ in range(8)
        ]
    ],
)
def test_step_async_pz_vector_env(use_single_action_space, env_fns):
    try:
        env_fns = [
            lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
            for _ in range(8)
        ]
        env = AsyncPettingZooVecEnv(env_fns)
        env.reset()
        if use_single_action_space:
            actions = {
                agent: [env.single_action_space(agent).sample() for _ in range(8)]
                for agent in env.agents
            }
        else:
            actions = {agent: env.action_space(agent).sample() for agent in env.agents}
        observations, rewards, terminations, truncations, _ = env.step(actions)
        for agent in env.agents:
            assert isinstance(env.single_action_space(agent), Discrete)
            assert isinstance(env.action_space(agent), MultiDiscrete)
            assert isinstance(env.observation_space(agent), Box)
            assert isinstance(observations[agent], np.ndarray)
            assert observations[agent].dtype == env.observation_space(agent).dtype
            assert (
                observations[agent].shape
                == (8,) + env.single_observation_space(agent).shape
            )
            assert observations[agent].shape == env.observation_space(agent).shape
            assert isinstance(rewards[agent], np.ndarray)
            assert isinstance(rewards[agent][0], (float, np.floating))
            assert rewards[agent].ndim == 1
            assert rewards[agent].size == 8
            assert isinstance(terminations[agent], np.ndarray)
            assert terminations[agent].dtype == np.bool_
            assert terminations[agent].ndim == 1
            assert terminations[agent].size == 8
            assert isinstance(truncations[agent], np.ndarray)
            assert truncations[agent].dtype == np.bool_
            assert truncations[agent].ndim == 1
            assert truncations[agent].size == 8
        env.close()

    except Exception as e:
        env.close()
        raise e


@pytest.mark.parametrize(
    "env_fns",
    [
        [
            lambda: simple_speaker_listener_v4.parallel_env(
                render_mode="rgb_array", continuous_actions=False
            )
            for _ in range(4)
        ]
    ],
)
def test_call_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.reset()

    images = env.call("render")
    max_num_agents = env.call("max_num_agents")
    env.close()

    assert isinstance(images, tuple)
    assert len(images) == 4
    for i in range(4):
        assert images[i].shape[-1] == 3
        assert isinstance(images[i][0], np.ndarray)

    assert isinstance(max_num_agents, tuple)
    assert len(max_num_agents) == 4
    for i in range(4):
        assert isinstance(max_num_agents[i], int)
        assert max_num_agents[i] == 2


@pytest.mark.parametrize(
    "env_fns",
    [
        [
            lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
            for _ in range(2)
        ]
    ],
)
def test_get_attr_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.set_attr("test_attribute", [1, 2])
    test_attribute = env.get_attr("test_attribute")
    assert test_attribute == (1, 2)
    env.close()


@pytest.mark.parametrize(
    "env_fns",
    [
        [
            lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
            for _ in range(1)
        ]
    ],
)
def test_set_attr_make_values_list(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.set_attr(name="test", values=1)
    assert env.call("test")[0] == 1
    env.close()


def test_env_order_preserved():
    env_fns = [
        lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
        for _ in range(16)
    ]
    env = AsyncPettingZooVecEnv(env_fns)
    env.reset()
    for obs_view in env.observations.obs_view.values():
        obs_view[:] = 0
    actions = {
        agent_id: env.single_action_space(agent_id).sample()
        for agent_id in env.possible_agents
    }
    rand_env = np.random.randint(0, 16)
    env.parent_pipes[rand_env].send(("step", actions))
    env.parent_pipes[rand_env].recv()
    for agent in env.agents:
        size = int(np.prod(env.single_observation_space(agent).shape))
        assert not np.array_equal(
            env.observations.obs_view[agent][rand_env * size : (rand_env + 1) * size],
            np.zeros_like(env.single_observation_space(agent).shape),
        )
    env.close()


def raise_error_reset(self, seed=None, options=None):
    if seed == 1:
        raise ValueError("Error in reset")
    return {
        agent: self.observation_space(agent).sample() for agent in self.possible_agents
    }, {agent: {} for agent in self.possible_agents}


def raise_error_step(self, action):
    if list(action.values())[0] >= 1:
        raise ValueError(f"Error in step with {action}")

    def pz_dict(transition, agents):
        return {agent: transition for agent in self.possible_agents}

    return (
        {
            agent: self.observation_space(agent).sample()
            for agent in self.possible_agents
        },
        pz_dict(0, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict({}, self.possible_agents),
    )


def test_async_vector_subenv_error():
    env_list = [
        lambda: GenericTestEnv(reset_func=raise_error_reset, step_func=raise_error_step)
    ]
    envs = AsyncPettingZooVecEnv(env_list * 2)

    with pytest.raises(ValueError, match="Error in reset"):
        envs.reset(seed=[1, 0])

    envs.close()
    env_list = [
        lambda: GenericTestEnv(reset_func=raise_error_reset, step_func=raise_error_step)
    ]
    envs = AsyncPettingZooVecEnv(env_list * 3)

    envs.reset()
    with pytest.raises(ValueError, match="Error in step"):
        envs.step({"agent_0": [0, 1, 2]})

    envs.close()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_reset_async_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_RESET
    with pytest.raises(AlreadyPendingCallError):
        env.reset_async()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_reset_wait_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    with pytest.raises(NoAsyncCallError):
        env.reset_async()
        env._state = AsyncState.DEFAULT
        env.reset_wait()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_step_async_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_RESET
    with pytest.raises(AlreadyPendingCallError):
        env.step_async(actions=None)
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_step_wait_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.DEFAULT
    with pytest.raises(NoAsyncCallError):
        env.step_wait()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_call_async_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_CALL
    with pytest.raises(AlreadyPendingCallError):
        env.call_async("test")
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_call_wait_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.DEFAULT
    with pytest.raises(NoAsyncCallError):
        env.call_wait()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_call_exception_worker(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    with pytest.raises(ValueError):
        env.call("reset")
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_set_attr_val_error(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    with pytest.raises(ValueError):
        env.set_attr("test", values=[1, 2, 3])
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_set_attr_exception(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_CALL
    with pytest.raises(AlreadyPendingCallError):
        env.set_attr("test", values=[1, 2])
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_close_extras_warning(env_fns):
    env = AsyncPettingZooVecEnv(
        env_fns,
    )
    env.reset_async()
    env._state = AsyncState.WAITING_RESET
    with patch.object(gym.logger, "warn") as mock_logger_warn:
        env.close_extras(timeout=None)
        mock_logger_warn.assert_called_once
    env.close()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_close_extras_terminate(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.reset_async()
    env._state = AsyncState.WAITING_RESET
    env.close_extras(terminate=True)

    for p in env.processes:
        assert not p.is_alive()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_poll_pipe_envs(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.parent_pipes[0] = None
    result = env._poll_pipe_envs(timeout=1)
    assert not result
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_assert_is_running(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env.closed = True
    with pytest.raises(ClosedEnvironmentError):
        env._assert_is_running()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_step_wait_timeout_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_STEP
    with pytest.raises(mp.TimeoutError):
        env.parent_pipes[0] = None
        env.step_wait(timeout=1)
        env.close()
    env.__del__()


@pytest.mark.parametrize(
    "env_fns", [[lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]]
)
def test_call_wait_timeout_async_pz_vector_env(env_fns):
    env = AsyncPettingZooVecEnv(env_fns)
    env._state = AsyncState.WAITING_CALL
    with pytest.raises(mp.TimeoutError):
        env.parent_pipes[0] = None
        env.call_wait(timeout=1)
        env.close()
    env.__del__()


@pytest.mark.parametrize(
    "transition_name", ["reward", "truncated", "terminated", "info", "observation"]
)
def test_get_placeholder_value(transition_name):
    env_fns = [lambda: simple_speaker_listener_v4.parallel_env() for _ in range(2)]
    env = AsyncPettingZooVecEnv(env_fns)
    if transition_name != "observation":
        val = get_placeholder_value("agent", transition_name)
        if transition_name == "reward":
            assert np.isnan(val)
        if transition_name == "truncated":
            assert np.isnan(val)
        if transition_name == "terminated":
            assert np.isnan(val)
        if transition_name == "info":
            assert val == {}
    else:
        env.reset()
        output = get_placeholder_value(
            agent="speaker_0",
            transition_name=transition_name,
            obs_spaces=env._single_observation_spaces,
        )
        assert isinstance(output, np.ndarray)
    env.close()


def test_add_info_dictionaries():
    info_list = [
        {
            "agent_0": {
                "env_defined_actions": np.array([1, 2, 3]),
                "action_mask": np.array([1, 0]),
            },
            "other_agent_0": {},
        },
        {
            "agent_0": {},
            "other_agent_0": {
                "env_defined_actions": np.array([5, 6, 7]),
                "action_mask": np.array([0, 1]),
            },
        },
        {"agent_0": {}, "other_agent_0": {}},
    ]
    env_fns = [lambda: simple_speaker_listener_v4.parallel_env() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)
    vector_infos = {}
    for i, info in enumerate(info_list):
        vector_infos = env._add_info(vector_infos, info, i)

    assert np.all(
        vector_infos["agent_0"]["env_defined_actions"]
        == np.array([[1, 2, 3], [0, 0, 0], [0, 0, 0]])
    )
    assert np.all(
        vector_infos["agent_0"]["action_mask"] == np.array([[1, 0], [0, 0], [0, 0]])
    )
    assert np.all(
        vector_infos["other_agent_0"]["env_defined_actions"]
        == np.array([[0, 0, 0], [5, 6, 7], [0, 0, 0]])
    )
    assert np.all(
        vector_infos["other_agent_0"]["action_mask"]
        == np.array([[0, 0], [0, 1], [0, 0]])
    )
    env.close()


def test_add_info_int():
    info_list = [
        {"agent_0": 1.0},
        {"other_agent_0": 1},
    ]
    env_fns = [lambda: GenericTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)
    vector_infos = {"agent_0": {}}
    for i, info in enumerate(info_list):
        vector_infos = env._add_info(vector_infos, info, i)

    env.close()


def test_add_info_unknown_objects():
    info_list = [
        {"agent_0": "string"},
        {"other_agent_0": "string"},
        {"agent_2": None},
    ]
    env_fns = [lambda: GenericTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)
    vector_infos = {"agent_0": {}}
    for i, info in enumerate(info_list):
        vector_infos = env._add_info(vector_infos, info, i)
    env.close()


def test_worker_reset():
    env_fns = [lambda: simple_speaker_listener_v4.parallel_env() for _ in range(1)]
    env_fn = env_fns[0]
    env = env_fn()
    env.reset()
    vec_env = AsyncPettingZooVecEnv(env_fns)
    parent_pipe, child_pipe = mp.Pipe()
    queue = mp.Queue()
    p = Process(
        target=_async_worker,
        args=(
            0,
            CloudpickleWrapper(env_fn),
            child_pipe,
            parent_pipe,
            vec_env._obs_buffer,
            queue,
            vec_env.agents,
        ),
    )
    p.start()
    child_pipe.close()
    parent_pipe.send(("reset", {}))
    results, success = parent_pipe.recv()
    assert success
    assert len(results) == 2
    assert list(sorted(results.keys())) == sorted(env.aec_env.agents)
    parent_pipe.close()
    p.terminate()
    p.join()


def test_worker_step_simple():
    num_envs = 1
    env_fns = [
        lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=True)
        for _ in range(num_envs)
    ]

    vec_env = AsyncPettingZooVecEnv(env_fns)
    vec_env.reset()

    actions = {
        agent: vec_env.single_action_space(agent).sample() for agent in vec_env.agents
    }
    vec_env.close()
    parent_pipe, child_pipe = mp.Pipe()
    queue = mp.Queue()
    p = Process(
        target=_async_worker,
        args=(
            0,
            CloudpickleWrapper(env_fns[0]),
            child_pipe,
            parent_pipe,
            vec_env._obs_buffer,
            queue,
            vec_env.agents,
        ),
    )
    p.start()
    child_pipe.close()
    # Reset the environment before stepping
    parent_pipe.send(("reset", {}))
    parent_pipe.recv()
    time.sleep(1)

    parent_pipe.send(("step", actions))
    results, success = parent_pipe.recv()
    assert success
    for dic in results:
        assert list(sorted(dic.keys())) == sorted(vec_env.agents)

    rewards, term, trunc, _ = results

    # state check
    assert vec_env.observations["speaker_0"].shape == (num_envs,) + (3,)
    assert vec_env.observations["listener_0"].shape == (num_envs,) + (11,)
    assert isinstance(vec_env.observations["speaker_0"], np.ndarray)
    assert isinstance(vec_env.observations["listener_0"], np.ndarray)

    # rewards check
    assert isinstance(rewards["speaker_0"], float)
    assert isinstance(rewards["listener_0"], float)

    # term check
    assert isinstance(term["speaker_0"], bool)
    assert isinstance(term["listener_0"], bool)

    # trunc check
    assert isinstance(trunc["speaker_0"], bool)
    assert isinstance(trunc["listener_0"], bool)

    parent_pipe.close()
    p.terminate()
    p.join()


def test_worker_step_autoreset():
    num_envs = 1
    env_fns = [lambda: term_env() for _ in range(num_envs)]
    vec_env = AsyncPettingZooVecEnv(env_fns)
    vec_env.reset()
    actions = {
        agent: np.array([vec_env.single_action_space(agent).sample()])
        for agent in vec_env.agents
    }
    vec_env.close()
    parent_pipe, child_pipe = mp.Pipe()
    queue = mp.Queue()
    p = Process(
        target=_async_worker,
        args=(
            0,
            CloudpickleWrapper(env_fns[0]),
            child_pipe,
            parent_pipe,
            vec_env._obs_buffer,
            queue,
            vec_env.agents,
        ),
    )
    p.start()
    child_pipe.close()
    # Reset the environment before stepping
    parent_pipe.send(("reset", {}))
    parent_pipe.recv()
    time.sleep(1)
    parent_pipe.send(("step", actions))
    results, success = parent_pipe.recv()
    assert success
    for dic in results:
        assert list(sorted(dic.keys())) == sorted(vec_env.agents)

    # Send step again and autoreset should be True
    parent_pipe.send(("step", actions))
    parent_pipe.recv()
    parent_pipe.close()
    p.terminate()
    p.join()


def test_worker_runtime_error():
    num_envs = 1
    env_fns = [
        lambda: simple_speaker_listener_v4.parallel_env() for _ in range(num_envs)
    ]
    vec_env = AsyncPettingZooVecEnv(env_fns)
    env_fn = env_fns[0]
    env = env_fn()
    env.reset()
    parent_pipe, child_pipe = mp.Pipe()
    queue = mp.Queue()
    try:
        p = Process(
            target=_async_worker,
            args=(
                0,
                CloudpickleWrapper(env_fn),
                child_pipe,
                parent_pipe,
                vec_env._obs_buffer,
                queue,
                vec_env.agents,
            ),
        )
        p.start()
        child_pipe.close()
        parent_pipe.send(("Unknown", {}))
        _, success = parent_pipe.recv()
        assert not success
        _, exctype, value, _ = queue.get()
        assert exctype.__name__ == "RuntimeError"
        assert isinstance(value, RuntimeError)
        assert (
            str(value)
            == "Received unknown command `Unknown`. Must be one of [`reset`, `step`, `close`, `_call`, `_setattr`, `_check_spaces`]."
        )
    finally:
        # Clean up resources
        os.kill(p.pid, signal.SIGTERM)
        p.join(timeout=1)  # Wait for process to terminate
        if p.is_alive():
            p.kill()  # Force kill if still alive
        p.join()  # Final join

        # Close pipes and queue
        parent_pipe.close()
        queue.close()
        queue.join_thread()


def test_observations_vector():
    num_envs = 1
    agents = ["speaker_0", "listener_0"]
    env_fns = [
        lambda: simple_speaker_listener_v4.parallel_env() for _ in range(num_envs)
    ]
    vec_env = AsyncPettingZooVecEnv(env_fns)
    assert (
        vec_env.observations.__str__()
        == vec_env.observations.__repr__()
        == "{'speaker_0': array([0., 0., 0.], dtype=float32), 'listener_0': array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.], dtype=float32)}"
    ), vec_env.observations.__str__()
    ob = {"speaker_0": np.ones((1, 3)), "listener_0": np.ones((1, 11))}
    write_to_shared_memory(
        0, ob, vec_env._obs_buffer, vec_env._single_observation_spaces
    )
    assert "speaker_0" in vec_env.observations
    assert len(vec_env.observations) == 2
    keys = []
    vals = []
    for key, val in vec_env.observations.items():
        keys.append(key)
        vals.append(val)
    assert keys == ["speaker_0", "listener_0"]
    assert np.all(np.concatenate(vals, axis=1) == np.ones((1, 14), dtype=np.float32))
    assert list(vec_env.observations.keys()) == agents
    assert np.all(
        np.concatenate(list(vec_env.observations.values()), axis=1)
        == np.ones((1, 14), dtype=np.float32)
    )
    assert np.all(
        vec_env.observations.get("speaker_0") == np.ones((1, 3), dtype=np.float32)
    )
    assert vec_env.observations.get("agent") is None
    assert (
        str(next(iter(vec_env.observations)))
        == "('speaker_0', array([[1., 1., 1.]], dtype=float32))"
    )
    vec_env.close()


def test_observations_image():
    num_envs = 1
    env_fns = [lambda: pursuit_v4.parallel_env() for _ in range(num_envs)]
    vec_env = AsyncPettingZooVecEnv(env_fns)

    # Reset the environment to initialize observations
    vec_env.reset()

    # Create a test observation with known values
    test_obs = {
        agent: np.ones((1, 7, 7, 3), dtype=np.uint8) * 255 for agent in vec_env.agents
    }

    # Write the test observation to shared memory
    write_to_shared_memory(
        0, test_obs, vec_env._obs_buffer, vec_env._single_observation_spaces
    )

    # Check if the values were correctly written to shared memory
    for agent in vec_env.agents:
        assert isinstance(vec_env.observations[agent], np.ndarray)
        assert vec_env.observations[agent].shape == (1, 7, 7, 3)
        assert np.all(vec_env.observations[agent] == 255)

    vec_env.close()


def test_write_to_shared_memory_dict_image():
    """Test writing dictionary observations with images to shared memory"""
    env_fns = [lambda: ComplexDictSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Create test observation
    test_obs = {
        "agent_0": {
            "position": np.ones((3,), dtype=np.float32),
            "velocity": np.ones((2,), dtype=np.float32) * 2,
            "image": np.ones((16, 16, 3), dtype=np.uint8) * 100,
        },
        "other_agent_0": {
            "position": np.ones((3,), dtype=np.float32) * 3,
            "velocity": np.ones((2,), dtype=np.float32) * 4,
            "image": np.ones((16, 16, 3), dtype=np.uint8) * 200,
        },
    }

    # Write to shared memory
    write_to_shared_memory(0, test_obs, env._obs_buffer, env._single_observation_spaces)

    # Check if values were correctly written
    for agent in env.agents:
        assert np.allclose(
            env.observations[agent]["position"][0], test_obs[agent]["position"]
        )
        assert np.allclose(
            env.observations[agent]["velocity"][0], test_obs[agent]["velocity"]
        )
        assert np.all(env.observations[agent]["image"][0] == test_obs[agent]["image"])

    env.close()


def test_write_to_shared_memory_tuple_image():
    """Test writing tuple observations with images to shared memory"""
    env_fns = [lambda: ComplexTupleSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Create test observation
    test_obs = {
        "agent_0": (
            np.ones((3,), dtype=np.float32),
            np.ones((2,), dtype=np.float32) * 2,
            np.ones((16, 16, 3), dtype=np.uint8) * 100,
        ),
        "other_agent_0": (
            np.ones((3,), dtype=np.float32) * 3,
            np.ones((2,), dtype=np.float32) * 4,
            np.ones((16, 16, 3), dtype=np.uint8) * 200,
        ),
    }

    # Write to shared memory
    write_to_shared_memory(0, test_obs, env._obs_buffer, env._single_observation_spaces)

    # Check if values were correctly written
    for agent in env.agents:
        assert np.allclose(env.observations[agent][0][0], test_obs[agent][0])
        assert np.allclose(env.observations[agent][1][0], test_obs[agent][1])
        assert np.all(env.observations[agent][2][0] == test_obs[agent][2])

    env.close()


dummy_action_spaces = {"agent_0": gym.spaces.Box(0, 1, (4,))}
dummy_observation_spaces = {"agent_0": gym.spaces.Box(0, 1, (4,))}


# Test for pz_vec_env.py
def test_vec_env_reset():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    vec_env.reset()


def test_vec_env_step():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    vec_env.step_async([])
    vec_env.step_wait()


def test_vec_env_render():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    with pytest.raises(NotImplementedError):
        vec_env.render()


def test_vec_env_closed():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    vec_env.closed = True
    vec_env.close()


def test_vec_env_close_extras():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    vec_env.close_extras()


def test_vec_env_unwrapped():
    vec_env = PettingZooVecEnv(
        3, dummy_observation_spaces, dummy_action_spaces, ["agent_0"]
    )
    vec_env.unwrapped


def test_delete_async_pz_vec_env():
    env_fns = [
        lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=False)
        for _ in range(2)
    ]
    env = AsyncPettingZooVecEnv(env_fns)
    assert len(env.processes) > 0  # Ensure subprocesses were created
    for process in env.processes:
        assert process.is_alive()
    processes = env.processes
    env.__del__()
    for p in processes:
        assert not p.is_alive()


def test_dict_space_env():
    """Test environment with dictionary observation spaces"""
    env_fns = [lambda: DictSpaceTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Check spaces
    for agent in env.agents:
        assert isinstance(env.single_observation_space(agent), spaces.Dict)
        assert isinstance(
            env.single_observation_space(agent).spaces["position"], spaces.Box
        )
        assert isinstance(
            env.single_observation_space(agent).spaces["velocity"], spaces.Box
        )

    # Test reset
    observations, infos = env.reset()
    for agent in env.agents:
        assert isinstance(observations[agent], dict)
        assert "position" in observations[agent]
        assert "velocity" in observations[agent]
        assert observations[agent]["position"].shape == (3, 3)
        assert observations[agent]["velocity"].shape == (3, 2)

    # Test step
    actions = {agent: env.action_space(agent).sample() for agent in env.agents}
    observations, rewards, terminations, truncations, infos = env.step(actions)
    for agent in env.agents:
        assert isinstance(observations[agent], dict)
        assert "position" in observations[agent]
        assert "velocity" in observations[agent]
        assert observations[agent]["position"].shape == (3, 3)
        assert observations[agent]["velocity"].shape == (3, 2)
        assert rewards[agent].shape == (3,)

    env.close()


def test_tuple_space_env():
    """Test environment with tuple observation spaces"""
    env_fns = [lambda: TupleSpaceTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Check spaces
    for agent in env.agents:
        assert isinstance(env.single_observation_space(agent), spaces.Tuple)
        assert isinstance(env.single_observation_space(agent).spaces[0], spaces.Box)
        assert isinstance(env.single_observation_space(agent).spaces[1], spaces.Box)

    # Test reset
    observations, infos = env.reset()
    for agent in env.agents:
        assert isinstance(observations[agent], tuple)
        assert len(observations[agent]) == 2
        assert observations[agent][0].shape == (3, 3)
        assert observations[agent][1].shape == (3, 2)

    # Test step
    actions = {agent: env.action_space(agent).sample() for agent in env.agents}
    observations, rewards, terminations, truncations, infos = env.step(actions)
    for agent in env.agents:
        assert isinstance(observations[agent], tuple)
        assert len(observations[agent]) == 2
        assert observations[agent][0].shape == (3, 3)
        assert observations[agent][1].shape == (3, 2)
        assert rewards[agent].shape == (3,)

    env.close()


def test_complex_dict_space_env():
    """Test environment with complex dictionary observation spaces (containing images)"""
    env_fns = [lambda: ComplexDictSpaceTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Check spaces
    for agent in env.agents:
        assert isinstance(env.single_observation_space(agent), spaces.Dict)
        assert isinstance(
            env.single_observation_space(agent).spaces["position"], spaces.Box
        )
        assert isinstance(
            env.single_observation_space(agent).spaces["velocity"], spaces.Box
        )
        assert isinstance(
            env.single_observation_space(agent).spaces["image"], spaces.Box
        )

    # Test reset
    observations, infos = env.reset()
    for agent in env.agents:
        assert isinstance(observations[agent], dict)
        assert "position" in observations[agent]
        assert "velocity" in observations[agent]
        assert "image" in observations[agent]
        assert observations[agent]["position"].shape == (3, 3)
        assert observations[agent]["velocity"].shape == (3, 2)
        assert observations[agent]["image"].shape == (3, 16, 16, 3)

    # Test step
    actions = {agent: env.action_space(agent).sample() for agent in env.agents}
    observations, rewards, terminations, truncations, infos = env.step(actions)
    for agent in env.agents:
        assert isinstance(observations[agent], dict)
        assert "position" in observations[agent]
        assert "velocity" in observations[agent]
        assert "image" in observations[agent]
        assert observations[agent]["position"].shape == (3, 3)
        assert observations[agent]["velocity"].shape == (3, 2)
        assert observations[agent]["image"].shape == (3, 16, 16, 3)
        assert rewards[agent].shape == (3,)

    env.close()


def test_complex_tuple_space_env():
    """Test environment with complex tuple observation spaces (containing images)"""
    env_fns = [lambda: ComplexTupleSpaceTestEnv() for _ in range(3)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Check spaces
    for agent in env.agents:
        assert isinstance(env.single_observation_space(agent), spaces.Tuple)
        assert isinstance(env.single_observation_space(agent).spaces[0], spaces.Box)
        assert isinstance(env.single_observation_space(agent).spaces[1], spaces.Box)
        assert isinstance(env.single_observation_space(agent).spaces[2], spaces.Box)

    # Test reset
    observations, infos = env.reset()
    for agent in env.agents:
        assert isinstance(observations[agent], tuple)
        assert len(observations[agent]) == 3
        assert observations[agent][0].shape == (3, 3)
        assert observations[agent][1].shape == (3, 2)
        assert observations[agent][2].shape == (3, 16, 16, 3)

    # Test step
    actions = {agent: env.action_space(agent).sample() for agent in env.agents}
    observations, rewards, terminations, truncations, infos = env.step(actions)
    for agent in env.agents:
        assert isinstance(observations[agent], tuple)
        assert len(observations[agent]) == 3
        assert observations[agent][0].shape == (3, 3)
        assert observations[agent][1].shape == (3, 2)
        assert observations[agent][2].shape == (3, 16, 16, 3)
        assert rewards[agent].shape == (3,)

    env.close()


def test_write_to_shared_memory_dict():
    """Test writing dictionary observations to shared memory"""
    env_fns = [lambda: DictSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Create test observation
    test_obs = {
        "agent_0": {
            "position": np.ones((3,), dtype=np.float32),
            "velocity": np.ones((2,), dtype=np.float32) * 2,
        },
        "other_agent_0": {
            "position": np.ones((3,), dtype=np.float32) * 3,
            "velocity": np.ones((2,), dtype=np.float32) * 4,
        },
    }

    # Write to shared memory
    write_to_shared_memory(0, test_obs, env._obs_buffer, env._single_observation_spaces)

    # Check if values were correctly written
    for agent in env.agents:
        assert np.allclose(
            env.observations[agent]["position"][0], test_obs[agent]["position"]
        )
        assert np.allclose(
            env.observations[agent]["velocity"][0], test_obs[agent]["velocity"]
        )

    env.close()


def test_write_to_shared_memory_tuple():
    """Test writing tuple observations to shared memory"""
    env_fns = [lambda: TupleSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    # Create test observation
    test_obs = {
        "agent_0": (
            np.ones((3,), dtype=np.float32),
            np.ones((2,), dtype=np.float32) * 2,
        ),
        "other_agent_0": (
            np.ones((3,), dtype=np.float32) * 3,
            np.ones((2,), dtype=np.float32) * 4,
        ),
    }

    # Write to shared memory
    write_to_shared_memory(0, test_obs, env._obs_buffer, env._single_observation_spaces)

    # Check if values were correctly written
    for agent in env.agents:
        assert np.allclose(env.observations[agent][0][0], test_obs[agent][0])
        assert np.allclose(env.observations[agent][1][0], test_obs[agent][1])

    env.close()


def test_placeholder_dict_space():
    """Test placeholder values for dictionary observation spaces"""
    env_fns = [lambda: DictSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    placeholder = get_placeholder_value(
        agent="agent_0",
        transition_name="observation",
        obs_spaces=env._single_observation_spaces,
    )

    assert isinstance(placeholder, dict)
    assert "position" in placeholder
    assert "velocity" in placeholder
    assert placeholder["position"].shape == (3,)
    assert placeholder["velocity"].shape == (2,)

    env.close()


def test_placeholder_tuple_space():
    """Test placeholder values for tuple observation spaces"""
    env_fns = [lambda: TupleSpaceTestEnv() for _ in range(1)]
    env = AsyncPettingZooVecEnv(env_fns)

    placeholder = get_placeholder_value(
        agent="agent_0",
        transition_name="observation",
        obs_spaces=env._single_observation_spaces,
    )

    assert isinstance(placeholder, tuple)
    assert len(placeholder) == 2
    assert placeholder[0].shape == (3,)
    assert placeholder[1].shape == (2,)

    env.close()


# Helper function to create a replay buffer and add transitions
def create_replay_buffer_with_transitions(env, memory_size=10):
    buffer = MultiAgentReplayBuffer(
        memory_size=memory_size,
        field_names=["state", "action", "reward", "next_state", "done"],
        agent_ids=env.possible_agents,
    )
    env.reset()
    for _ in range(memory_size):
        actions = {
            agent: env.action_space(agent).sample() for agent in env.possible_agents
        }
        obs, rewards, dones, truncated, infos = env.step(actions)
        buffer.save_to_memory(obs, actions, rewards, obs, dones, is_vectorised=True)
    return buffer


@pytest.mark.parametrize("env_cls", [DictSpaceTestEnv, TupleSpaceTestEnv])
def test_replay_buffer_with_various_spaces(env_cls):
    env_fns = [lambda: env_cls() for _ in range(3)]  # 3 parallel environments
    env = AsyncPettingZooVecEnv(env_fns)

    buffer = create_replay_buffer_with_transitions(env)
    batch_size = 5
    sampled_transitions = buffer.sample(batch_size)

    # Check that the sampled transitions have the correct structure
    for field, agent_data in zip(buffer.field_names, sampled_transitions):
        for agent_id, data in agent_data.items():
            assert agent_id in env.possible_agents
            if field == "state":
                obs_space = env.single_observation_space(agent_id)
                if isinstance(obs_space, spaces.Dict):
                    for key in obs_space.spaces:
                        assert key in data
                        assert data[key].shape[0] == batch_size
                elif isinstance(obs_space, spaces.Tuple):
                    assert len(data) == len(obs_space.spaces)
                    for i, sub_data in enumerate(data):
                        assert sub_data.shape[0] == batch_size
                else:
                    assert data.shape[0] == batch_size

    env.close()



from __future__ import annotations

import os
import pickle as pkl
from typing import Any


class Cache:
    def __init__(self, cache_init: dict | None = None) -> None:
        """Initializes the Cache object.

        :param cache_init: Initial cache dictionary.
        :type cache_init: dict | None
        """
        assert cache_init is None or isinstance(cache_init, dict)
        if cache_init is None:
            cache_init = {}
        self.cache = cache_init
        self.cache_hit_rate = 1.0

    def dump(self, file_name: str) -> None:
        """Dumps the cache to a file.

        :param file_name: Name of the file to dump the cache to.
        :type file_name: str
        """
        if not os.path.exists(os.path.dirname(file_name)):
            os.makedirs(os.path.dirname(file_name))
        with open(file_name, "wb") as f:
            pkl.dump(self.cache, f)

    def load(self, file_name: str) -> None:
        """Loads the cache from a file.

        :param file_name: Name of the file to load the cache from.
        :type file_name: str
        """
        with open(file_name, "rb") as f:
            self.cache.update(pkl.load(f))

    def __getitem__(self, key: str) -> Any:
        """Gets an item from the cache.

        :param key: Key of the item to get.
        :type key: str

        :return: Item from the cache.
        :rtype: Any
        """
        self.cache_hit_rate = (self.cache_hit_rate * 0.99) + 0.01
        return self.cache[key]

    def __setitem__(self, key: str, newvalue: Any) -> None:
        """Sets an item in the cache.

        :param key: Key of the item to set.
        :type key: str
        :param newvalue: Value to set.
        :type newvalue: Any
        """
        self.cache_hit_rate = self.cache_hit_rate * 0.99
        self.cache[key] = newvalue

    def __contains__(self, key: str) -> bool:
        """Checks if a key is in the cache.

        :param key: Key to check.
        :type key: str

        :return: True if the key is in the cache, False otherwise.
        :rtype: bool
        """
        return key in self.cache

    def __len__(self) -> int:
        """Gets the length of the cache.

        :return: Length of the cache.
        :rtype: int
        """
        return len(self.cache)

    def items(self) -> list[tuple[str, Any]]:
        """Gets the items of the cache.

        :return: Items of the cache.
        :rtype: list
        """
        return list(self.cache.items())

    def keys(self) -> list[str]:
        """Gets the keys of the cache.

        :return: Keys of the cache.
        :rtype: list
        """
        return list(self.cache.keys())

    def values(self) -> list[Any]:
        """Gets the values of the cache.

        :return: Values of the cache.
        :rtype: list
        """
        return list(self.cache.values())

    def update(self, new_stuff: dict) -> None:
        """Updates the cache with new items.

        :param new_stuff: New items to update the cache with.
        :type new_stuff: dict
        """
        self.cache.update(new_stuff)

    def get_hit_rate(self) -> float:
        """Gets the hit rate of the cache.

        :return: Hit rate of the cache.
        :rtype: float
        """
        return self.cache_hit_rate

    def get_cache(self) -> dict:
        """Gets the cache.

        :return: Cache.
        :rtype: dict
        """
        return self.cache



from typing import Any, Optional

import torch
import torch.nn as nn

from agilerl.modules import EvolvableCNN
from agilerl.modules.base import EvolvableModule, MutationType, mutation
from agilerl.typing import DeviceType, ObservationType
from agilerl.utils.evolvable_networks import create_resnet, get_activation


class EvolvableResNet(EvolvableModule):
    """Evolvable module that implements the architecture presented in 'Deep Residual Learning
    for Image Recognition'. Designed to train CNN's more successfully by introducing residual
    connections that skip one or more layers.

    Paper: https://arxiv.org/abs/1512.03385

    :param input_shape: Input shape of the neural network
    :type input_shape: list[int]
    :param num_outputs: Output layer dimension
    :type num_outputs: int
    :param channel_size: A list of integers representing the number of channels in each convolutional layer
    :type channel_size: int
    :param kernel_size: A list of integers representing the kernel size of each convolutional layer
    :type kernel_size: int
    :param stride_size: A list of integers representing the stride size of each convolutional layer
    :type stride_size: int
    :param num_blocks: Number of residual blocks that compose the network
    :type num_blocks: int
    :param output_activation: Output activation layer, defaults to None
    :type output_activation: str, optional
    :param scale_factor: Scale factor for the network, defaults to 4
    :type scale_factor: int
    :param min_blocks: Minimum number of residual blocks that compose the network, defaults to 1
    :type min_blocks: int
    :param max_blocks: Maximum number of residual blocks that compose the network, defaults to 4
    :type max_blocks: int
    :param min_channel_size: Minimum number of channels in each convolutional layer, defaults to 32
    :type min_channel_size: int
    :param max_channel_size: Maximum number of channels in each convolutional layer, defaults to 256
    :type max_channel_size: int
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param name: Name of the network, defaults to 'resnet'
    :type name: str, optional
    :param random_seed: Random seed to use for the network. Defaults to None.
    :type random_seed: Optional[int]
    """

    def __init__(
        self,
        input_shape: list[int],
        num_outputs: int,
        channel_size: int,
        kernel_size: int,
        stride_size: int,
        num_blocks: int,
        output_activation: str = None,
        scale_factor: int = 4,
        min_blocks: int = 1,
        max_blocks: int = 4,
        min_channel_size: int = 32,
        max_channel_size: int = 256,
        device: DeviceType = "cpu",
        name: str = "resnet",
        random_seed: Optional[int] = None,
    ) -> None:
        super().__init__(device, random_seed)

        assert isinstance(scale_factor, int), "Scale factor must be an integer."
        assert isinstance(num_blocks, int), "Number of blocks must be an integer."
        assert isinstance(channel_size, int), "Channel size must be an integer."
        assert num_outputs >= 1, "Number of blocks must be greater than or equal to 1."
        assert num_blocks >= 1, "Number of blocks must be greater than or equal to 1."

        self.input_shape = input_shape
        self.num_outputs = num_outputs
        self.channel_size = channel_size
        self.kernel_size = kernel_size
        self.stride_size = stride_size
        self.num_blocks = num_blocks
        self.output_activation = output_activation
        self.scale_factor = scale_factor
        self.min_blocks = min_blocks
        self.max_blocks = max_blocks
        self.min_channel_size = min_channel_size
        self.max_channel_size = max_channel_size
        self.name = name

        self.model = self.create_resnet(
            input_shape=self.input_shape,
            channel_size=channel_size,
            kernel_size=kernel_size,
            stride_size=stride_size,
            num_blocks=num_blocks,
            scale_factor=self.scale_factor,
        )

    @property
    def net_config(self) -> dict[str, Any]:
        """Returns model configuration in dictionary."""
        net_config = self.init_dict.copy()
        for attr in ["num_inputs", "num_outputs", "device", "name"]:
            if attr in net_config:
                net_config.pop(attr)

        return net_config

    def change_activation(self, activation: str, output: bool = False) -> None:
        """We currently do not support changing the activation function of the ResNet.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional
        """
        return

    def create_resnet(
        self,
        input_shape: list[int],
        channel_size: int,
        kernel_size: int,
        stride_size: int,
        num_blocks: int,
        scale_factor: int,
    ) -> nn.Sequential:
        """
        Creates and returns a convolutional neural network.

        :param in_channels: The number of input channels.
        :type in_channels: int
        :param channel_size: A list of integers representing the number of channels in each convolutional layer.
        :type channel_size: list[int]
        :param kernel_size: A list of integers representing the kernel size of each convolutional layer.
        :type kernel_size: list[int]
        :param stride_size: A list of integers representing the stride size of each convolutional layer.
        :type stride_size: list[int]
        :param num_blocks: The number of residual blocks in the CNN.
        :type num_blocks: int
        :param scale_factor: The scale factor for the CNN.
        :type scale_factor: int

        :return: The created convolutional neural network.
        :rtype: nn.Sequential
        """
        # Build the main convolutional block
        net_dict = create_resnet(
            input_channels=input_shape[0],
            channel_size=channel_size,
            kernel_size=kernel_size,
            stride_size=stride_size,
            num_blocks=num_blocks,
            scale_factor=scale_factor,
            device=self.device,
            name=self.name,
        )

        # Flatten image encodings and pass through a final linear layer
        sample_input = torch.zeros(1, *input_shape, device=self.device)
        pre_flatten_output = nn.Sequential(net_dict)(sample_input)
        net_dict[f"{self.name}_flatten"] = nn.Flatten()
        with torch.no_grad():
            cnn_output = nn.Sequential(net_dict)(sample_input)
            flattened_size = cnn_output.shape[1]

        net_dict[f"{self.name}_linear_output"] = nn.Linear(
            flattened_size, self.num_outputs, device=self.device
        )
        net_dict[f"{self.name}_output_activation"] = get_activation(
            self.output_activation
        )

        self.cnn_output_size = pre_flatten_output.shape

        return nn.Sequential(net_dict)

    def forward(self, x: ObservationType) -> torch.Tensor:
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor
        :return: Neural network output
        :rtype: torch.Tensor
        """
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, dtype=torch.float32, device=self.device)

        if len(x.shape) == 3:
            x = x.unsqueeze(0)

        return self.model(x)

    @mutation(MutationType.LAYER)
    def add_block(self) -> None:
        """Adds a hidden layer to neural network. Falls back on ``add_channel()`` if
        max hidden layers reached."""
        # add layer to hyper params
        if self.num_blocks < self.max_blocks:  # HARD LIMIT
            self.num_blocks += 1
        else:
            return self.add_channel()

    @mutation(MutationType.LAYER, shrink_params=True)
    def remove_block(self) -> None:
        """Removes a hidden layer from neural network. Falls back on ``add_channel()`` if
        min hidden layers reached."""
        if self.num_blocks > self.min_blocks:  # HARD LIMIT
            self.num_blocks -= 1
        else:
            return self.add_channel()

    @mutation(MutationType.NODE)
    def add_channel(
        self,
        numb_new_channels: Optional[int] = None,
    ) -> dict[str, int]:
        """Remove channel from hidden layer of convolutional neural network.

        :param numb_new_channels: Number of channels to add to hidden layer, defaults to None
        :type numb_new_channels: int, optional
        :return: Dictionary containing the hidden layer and number of new channels
        :rtype: dict[str, Union[int, None]]
        """
        if numb_new_channels is None:
            numb_new_channels = self.rng.choice([8, 16, 32])

        # HARD LIMIT
        if self.channel_size + numb_new_channels < self.max_channel_size:
            self.channel_size += numb_new_channels

        return {"numb_new_channels": numb_new_channels}

    @mutation(MutationType.NODE, shrink_params=True)
    def remove_channel(
        self,
        numb_new_channels: Optional[int] = None,
    ) -> dict[str, int]:
        """Remove channel from hidden layer of convolutional neural network.

        :param numb_new_channels: Number of channels to add to hidden layer, defaults to None
        :type numb_new_channels: int, optional
        :return: Dictionary containing the hidden layer and number of new channels
        :rtype: dict[str, Union[int, None]]
        """
        if numb_new_channels is None:
            numb_new_channels = self.rng.choice([8, 16, 32])

        # HARD LIMIT
        if self.channel_size - numb_new_channels > self.min_channel_size:
            self.channel_size -= numb_new_channels

        return {"numb_new_channels": numb_new_channels}

    def recreate_network(self, shrink_params: bool = False) -> None:
        """Recreates neural networks.

        :param shrink_params: Flag indicating whether to shrink the parameters, defaults to False
        :type shrink_params: bool, optional
        """
        # Create model with new architecture
        model = self.create_resnet(
            input_shape=self.input_shape,
            channel_size=self.channel_size,
            kernel_size=self.kernel_size,
            stride_size=self.stride_size,
            num_blocks=self.num_blocks,
            scale_factor=self.scale_factor,
        )

        # Copy parameters from old model to new model
        preserve_params_fn = (
            EvolvableCNN.shrink_preserve_parameters
            if shrink_params
            else EvolvableModule.preserve_parameters
        )
        self.model = preserve_params_fn(old_net=self.model, new_net=model)



import types
from collections.abc import Callable

import gymnasium as gym
import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from pettingzoo import ParallelEnv

ROCK = 0
PAPER = 1
SCISSORS = 2
NONE = 3
MOVES = ["ROCK", "PAPER", "SCISSORS", "None"]
NUM_ITERS = 100
REWARD_MAP = {
    (ROCK, ROCK): (0, 0),
    (ROCK, PAPER): (-1, 1),
    (ROCK, SCISSORS): (1, -1),
    (PAPER, ROCK): (1, -1),
    (PAPER, PAPER): (0, 0),
    (PAPER, SCISSORS): (-1, 1),
    (SCISSORS, ROCK): (-1, 1),
    (SCISSORS, PAPER): (1, -1),
    (SCISSORS, SCISSORS): (0, 0),
}


class term_env(ParallelEnv):
    metadata = {"render_modes": ["human"], "name": "rps_v2"}

    def __init__(self, render_mode=None):
        self.possible_agents = ["player_" + str(r) for r in range(2)]
        self.agent_name_mapping = dict(
            zip(self.possible_agents, list(range(len(self.possible_agents))))
        )
        self.render_mode = render_mode
        self.seed = self.dummy_seed

    def dummy_seed(self, seed):
        pass

    def observation_space(self, agent):
        return Discrete(2)  # 0, 2, shape=(2,), dtype=float)

    def action_space(self, agent):
        return Box(0, 2, shape=(1,), dtype=float)

    def close(self):
        pass

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents[:]
        self.num_moves = 0
        observations = {agent: NONE for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        self.state = observations
        return observations, infos

    def step(self, actions):
        if not actions:
            self.agents = []
            return {}, {}, {}, {}, {}
        rewards = {}
        rewards[self.agents[0]], rewards[self.agents[1]] = (0, 0)
        terminations = {agent: True for agent in self.agents}
        self.num_moves += 1
        env_truncation = self.num_moves >= NUM_ITERS
        truncations = {agent: env_truncation for agent in self.agents}
        observations = {
            self.agents[i]: actions[self.agents[1 - i]].astype(int)
            for i in range(len(self.agents))
        }
        self.state = observations
        infos = {agent: {} for agent in self.agents}
        if env_truncation:
            self.agents = []
        return observations, rewards, terminations, truncations, infos


def basic_reset_func(
    self,
    *,
    seed: int | None = None,
    options: dict | None = None,
):
    """A basic reset function that will pass the environment check using random actions from the observation space."""
    return {
        agent: self.observation_space(agent).sample() for agent in self.possible_agents
    }, {agent: {"options": options} for agent in self.possible_agents}


def old_reset_func(self):
    """An old reset function that will pass the environment check using random actions from the observation space."""
    super(GenericTestEnv, self).reset()
    return {
        agent: self.observation_space(agent).sample() for agent in self.possible_agent
    }


def basic_step_func(self, action):
    """A step function that follows the basic step api that will pass the environment check using random actions from the observation space."""

    def pz_dict(transition, agents):
        return {agent: transition for agent in self.possible_agents}

    return (
        {
            agent: self.observation_space(agent).sample()
            for agent in self.possible_agents
        },
        pz_dict(0, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict({}, self.possible_agents),
    )


def old_step_func(self, action):
    """A step function that follows the old step api that will pass the environment check using random actions from the observation space."""

    def pz_dict(transition, agents):
        return {agent: transition for agent in self.possible_agents}

    return (
        {
            agent: self.observation_space(agent).sample()
            for agent in self.possible_agents
        },
        pz_dict(0, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict(False, self.possible_agents),
        pz_dict({}, self.possible_agents),
    )


def basic_render_func(self):
    """Basic render fn that does nothing."""
    pass


class GenericTestEnv(ParallelEnv):
    """A generic testing environment for use in testing with modified environments are required."""

    def __init__(
        self,
        action_space: spaces.Space = spaces.Box(0, 1, (1,)),
        observation_space: spaces.Space = spaces.Box(0, 1, (1,)),
        reset_func: Callable = basic_reset_func,
        step_func: Callable = basic_step_func,
        render_func: Callable = basic_render_func,
        metadata={"render_modes": []},
        render_mode=None,
    ):
        """Generic testing environment constructor.

        Args:
            action_space: The environment action space
            observation_space: The environment observation space
            reset_func: The environment reset function
            step_func: The environment step function
            render_func: The environment render function
            metadata: The environment metadata
            render_mode: The render mode of the environment
            spec: The environment spec
        """
        self.metadata = metadata
        self.render_mode = render_mode
        self.possible_agents = ["agent_0"]
        self.action_space_dict = {"agent_0": action_space}
        self.observation_space_dict = {"agent_0": observation_space}

        if observation_space is not None:
            self.observation_space = self.action_space_func
        if action_space is not None:
            self.action_space = self.observation_space_func

        if reset_func is not None:
            self.reset = types.MethodType(reset_func, self)
        if step_func is not None:
            self.step = types.MethodType(step_func, self)
        if render_func is not None:
            self.render = types.MethodType(render_func, self)

    def action_space_func(self, agent):
        return self.action_space_dict[agent]

    def observation_space_func(self, agent):
        return self.observation_space_dict[agent]

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict | None = None,
    ):
        """Resets the environment."""
        # If you need a default working reset function, use `basic_reset_fn` above
        raise NotImplementedError("TestingEnv reset_fn is not set.")

    def step(self, action):
        """Steps through the environment."""
        raise NotImplementedError("TestingEnv step_fn is not set.")

    def render(self):
        """Renders the environment."""
        raise NotImplementedError("testingEnv render_fn is not set.")


class CustomSpace(gym.Space):
    """Minimal custom observation space."""

    shape = (4,)
    _dtype = np.float16

    def sample(self):
        """Generates a sample from the custom space."""
        return self.np_random.integers(0, 10, ())

    def contains(self, x):
        """Check if the element `x` is contained within the space."""
        return 0 <= x <= 10

    def __eq__(self, other):
        """Check if the two spaces are equal."""
        return isinstance(other, CustomSpace)

    @property
    def dtype(self):
        return self._dtype



import os
from unittest.mock import MagicMock, Mock, patch

import gymnasium as gym
import numpy as np
import pytest
import torch
from accelerate import Accelerator, DeepSpeedPlugin
from gymnasium import spaces
from pettingzoo.mpe import simple_speaker_listener_v4

from agilerl.algorithms import (
    CQN,
    DDPG,
    DQN,
    GRPO,
    IPPO,
    MADDPG,
    MATD3,
    PPO,
    TD3,
    RainbowDQN,
)
from agilerl.algorithms.core import EvolvableAlgorithm, LLMAlgorithm
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.utils import (
    aggregate_metrics_across_gpus,
    calculate_vectorized_scores,
    consolidate_mutations,
    create_population,
    gather_tensor,
    make_multi_agent_vect_envs,
    make_skill_vect_envs,
    make_vect_envs,
    plot_population_score,
    print_hyperparams,
    save_llm_checkpoint,
    tournament_selection_and_mutation,
)
from agilerl.wrappers.learning import Skill

# Shared HP dict that can be used by any algorithm
SHARED_INIT_HP = {
    "POPULATION_SIZE": 4,
    "DOUBLE": True,
    "BATCH_SIZE": 128,
    "CUDAGRAPHS": False,
    "LR": 1e-3,
    "LR_ACTOR": 1e-4,
    "LR_CRITIC": 1e-3,
    "GAMMA": 0.99,
    "LEARN_STEP": 1,
    "TAU": 1e-3,
    "BETA": 0.4,
    "PRIOR_EPS": 0.000001,
    "NUM_ATOMS": 51,
    "V_MIN": 0,
    "V_MAX": 200,
    "N_STEP": 3,
    "POLICY_FREQ": 10,
    "GAE_LAMBDA": 0.95,
    "ACTION_STD_INIT": 0.6,
    "CLIP_COEF": 0.2,
    "ENT_COEF": 0.01,
    "VF_COEF": 0.5,
    "MAX_GRAD_NORM": 0.5,
    "TARGET_KL": None,
    "UPDATE_EPOCHS": 4,
    "AGENT_IDS": ["agent1", "agent2"],
    "LAMBDA": 1.0,
    "REG": 0.000625,
    "CHANNELS_LAST": False,
    "O_U_NOISE": True,
    "EXPL_NOISE": 0.1,
    "MEAN_NOISE": 0.0,
    "THETA": 0.15,
    "DT": 0.01,
}


# Returns an AsyncVectorEnv object when given a valid environment name and number of environments
def test_returns_asyncvectorenv_object():
    num_envs = 3
    env = make_vect_envs("CartPole-v1", num_envs=num_envs)
    assert isinstance(env, gym.vector.AsyncVectorEnv)
    assert env.num_envs == num_envs


# Returns an AsyncVectorEnv object when given a valid environment name and number of environments
def test_returns_asyncvectorenv_object_multiagent():
    num_envs = 3
    env = simple_speaker_listener_v4.parallel_env
    env_kwargs = {"continuous_actions": False}
    env = make_multi_agent_vect_envs(env, num_envs=num_envs, **env_kwargs)
    env.close()
    assert env.num_envs == num_envs


# Returns an AsyncVectorEnv object when given a valid environment name and number of environments
def test_returns_asyncvectorenv_object_skill():
    num_envs = 3
    skill = Skill
    env = make_skill_vect_envs("CartPole-v1", skill=skill, num_envs=num_envs)
    assert isinstance(env, gym.vector.AsyncVectorEnv)
    assert env.num_envs == num_envs


# Can create a population of agent for each single agent algorithm
def test_create_initial_population_single_agent():
    observation_space = spaces.Box(0, 1, shape=(4,))
    continuous_action_space = spaces.Box(0, 1, shape=(2,))
    discrete_action_space = spaces.Discrete(2)
    net_config = {"encoder_config": {"hidden_size": [8, 8]}}
    population_size = 4
    device = "cpu"
    accelerator = None

    algo_classes = {
        "DQN": DQN,
        "Rainbow DQN": RainbowDQN,
        "DDPG": DDPG,
        "TD3": TD3,
        "PPO": PPO,
        "CQN": CQN,
    }

    for algo in algo_classes.keys():
        if algo in ["TD3", "DDPG"]:
            action_space = continuous_action_space
        else:
            action_space = discrete_action_space

        population = create_population(
            algo=algo,
            observation_space=observation_space,
            action_space=action_space,
            net_config=net_config,
            INIT_HP=SHARED_INIT_HP,
            population_size=population_size,
            device=device,
            accelerator=accelerator,
        )
        assert len(population) == population_size
        for agent in population:
            assert isinstance(agent, algo_classes[algo])
            assert agent.observation_space == observation_space
            assert agent.action_space == action_space
            assert agent.device == "cpu"
            assert agent.accelerator is None


# Can create a population of agent for each multi agent algorithm
def test_create_initial_population_multi_agent():
    observation_space = [spaces.Box(0, 1, shape=(4,)) for _ in range(2)]
    action_space = [spaces.Discrete(2) for _ in range(2)]
    net_config = {"encoder_config": {"hidden_size": [8], "min_mlp_nodes": 2}}
    population_size = 4
    device = "cpu"
    accelerator = None

    algo_classes = {
        "MADDPG": MADDPG,
        "MATD3": MATD3,
        "IPPO": IPPO,
    }

    for algo in algo_classes.keys():
        population = create_population(
            algo=algo,
            observation_space=observation_space,
            action_space=action_space,
            net_config=net_config,
            INIT_HP=SHARED_INIT_HP,
            population_size=population_size,
            device=device,
            accelerator=accelerator,
        )
        assert len(population) == population_size
        for agent in population:
            assert isinstance(agent, algo_classes[algo])
            assert agent.observation_spaces == observation_space
            assert agent.action_spaces == action_space
            assert agent.device == "cpu"
            assert agent.accelerator is None


# The function returns a list of episode rewards from the first episode in each parallel environment.
def test_returns_list_of_episode_rewards():
    rewards = np.array([[1, 2, 3, 4, 5], [4, 5, 6, 7, 8]])
    terminations = np.array([[0, 0, 1, 0, 1], [0, 1, 0, 0, 0]])
    expected_rewards = [6, 9]

    result = calculate_vectorized_scores(
        rewards, terminations, include_unterminated=False, only_first_episode=True
    )

    assert result == expected_rewards


# The function returns a list of episode rewards including all episodes.
def test_returns_list_of_episode_rewards_including_unterminated():
    rewards = np.array([[1, 2, 3], [4, 5, 6]])
    terminations = np.array([[0, 0, 1], [0, 1, 0]])
    expected_rewards = [6, 9, 6]

    result = calculate_vectorized_scores(
        rewards, terminations, include_unterminated=True, only_first_episode=False
    )

    assert result == expected_rewards


# The function returns a list of episode rewards including all terminated episodes.
def test_returns_list_of_episode_rewards_all_terminated_episodes():
    rewards = np.array([[1, 2, 3, 4, 5], [4, 5, 6, 7, 8]])
    terminations = np.array([[0, 0, 1, 0, 1], [0, 1, 0, 0, 0]])
    expected_rewards = [6, 9, 9]

    result = calculate_vectorized_scores(
        rewards, terminations, include_unterminated=False, only_first_episode=False
    )

    assert result == expected_rewards


# The function returns a list of episode rewards including all terminated episodes.
def test_returns_list_of_episode_rewards_including_all_terminated_episodes():
    rewards = np.array([[1, 2, 3, 4, 5], [4, 5, 6, 7, 8]])
    terminations = np.array([[0, 0, 1, 0, 1], [0, 1, 0, 0, 0]])
    expected_rewards = [6, 9, 9]

    result = calculate_vectorized_scores(
        rewards, terminations, include_unterminated=False, only_first_episode=False
    )

    assert result == expected_rewards


# The function returns a list of episode rewards containing no terminated episodes.
def test_returns_list_of_episode_rewards_with_no_terminations():
    rewards = np.array([[1, 2, 3, 4, 5], [4, 5, 6, 7, 8]])
    terminations = np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
    expected_rewards = [15, 30]

    result = calculate_vectorized_scores(
        rewards, terminations, include_unterminated=True, only_first_episode=False
    )

    assert result == expected_rewards


# The function prints the hyperparameters and fitnesses of all agents in the population.
def test_prints_hyperparams():
    # Arrange
    observation_space = spaces.Box(0, 1, shape=(4,))
    action_space = spaces.Discrete(2)
    net_config = {"encoder_config": {"hidden_size": [8]}}
    population_size = 1
    device = "cpu"
    accelerator = None
    algo = "DQN"

    pop = create_population(
        algo=algo,
        observation_space=observation_space,
        action_space=action_space,
        net_config=net_config,
        INIT_HP=SHARED_INIT_HP,
        population_size=population_size,
        device=device,
        accelerator=accelerator,
    )

    # Manually set attributes
    pop[0].fitness = [1, 2, 3]
    pop[0].lr = 0.01
    pop[0].batch_size = 32

    expected_output = "Agent ID: {}    Mean 5 Fitness: {:.2f}    Attributes: {}".format(
        pop[0].index,
        np.mean(pop[0].fitness[-5:]),
        EvolvableAlgorithm.inspect_attributes(pop[0]),
    )

    with patch("builtins.print") as mock_print:
        print_hyperparams(pop)
        mock_print.assert_called_once_with(expected_output)


# The function should correctly plot the fitness scores of all agents in the population.
@patch("agilerl.utils.utils.plt")
def test_plot_fitness_scores_all_agents(mock_plt):
    # Create a population of agents with fitness scores
    class Agent:
        def __init__(self, fitness):
            self.fitness = fitness
            self.steps = list(range(len(fitness) + 1))

    pop = [Agent([1, 2, 3]), Agent([4, 5, 6]), Agent([7, 8, 9])]

    # Call the function under test
    plot_population_score(pop)

    # Assert plotting functions have been called with expected args
    mock_plt.title.assert_called_once_with("Score History - Mutations")
    mock_plt.xlabel.assert_called_once_with("Steps")

    # Assert plt.figure got called
    assert mock_plt.figure.called


def test_save_with_accelerator():
    """Test saving checkpoint when agent has an accelerator."""
    agent = Mock()
    agent.actor = Mock()
    agent.accelerator = Mock()
    agent.accelerator.wait_for_everyone = Mock()
    agent.algo = "grpo"
    save_llm_checkpoint(agent, None)
    agent.save_checkpoint.assert_called_once_with(
        "./saved_checkpoints/grpo", weights_only=False
    )
    agent.accelerator.wait_for_everyone.assert_called()
    os.rmdir("saved_checkpoints/grpo")


def test_save_without_accelerator():
    """Test saving checkpoint when agent has no accelerator."""
    agent = Mock()
    agent.actor = Mock()
    agent.algo = "grpo"
    agent.accelerator = None
    save_llm_checkpoint(agent, None)
    agent.save_checkpoint.assert_called_once_with(
        "./saved_checkpoints/grpo", weights_only=False
    )
    os.rmdir("saved_checkpoints/grpo")


def test_gather_tensor_with_tensor_input():
    """Test gather_tensor with tensor input"""

    accelerator = Accelerator()

    input_tensor = torch.tensor([1, 2, 3], device=accelerator.device)

    gathered = gather_tensor(input_tensor, accelerator)

    assert isinstance(gathered, torch.Tensor)

    assert torch.equal(gathered, input_tensor)


def test_gather_tensor_with_non_tensor_input():
    """Test gather_tensor with non-tensor input"""
    input_list = [1, 2, 3]

    accelerator = Accelerator()

    gathered = gather_tensor(input_list, accelerator)

    assert isinstance(gathered, torch.Tensor)

    assert torch.equal(gathered, torch.tensor(input_list).to(accelerator.device))


def test_gather_tensor_device():
    """Test that tensor is moved to accelerator device"""
    input_tensor = torch.tensor([1, 2, 3])

    accelerator = Accelerator()

    gathered = gather_tensor(input_tensor, accelerator)

    assert gathered.device.type == accelerator.device.type


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_gather_tensor_distributed():
    """Test gather_tensor in distributed setting"""

    accelerator = Accelerator()

    rank = accelerator.process_index
    input_tensor = torch.tensor([rank], device=accelerator.device)

    gathered = gather_tensor(input_tensor, accelerator)

    assert len(gathered) == accelerator.num_processes
    assert torch.equal(
        gathered, torch.arange(accelerator.num_processes, device=accelerator.device)
    )


def test_aggregate_metrics_across_gpus_single_process():
    """Test aggregate_metrics_across_gpus with single process"""
    accelerator = Accelerator()

    metric_tensor = torch.tensor([1.0, 2.0, 3.0], device=accelerator.device)

    result = aggregate_metrics_across_gpus(accelerator, metric_tensor)

    assert result == 2.0  # (1 + 2 + 3) / 3 = 2.0
    assert isinstance(result, float)


def test_aggregate_metrics_across_gpus_with_scalar():
    """Test aggregate_metrics_across_gpus with scalar input"""
    accelerator = Accelerator()

    metric_tensor = torch.tensor(5.0, device=accelerator.device)

    result = aggregate_metrics_across_gpus(accelerator, metric_tensor)

    assert result == 5.0
    assert isinstance(result, float)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_aggregate_metrics_across_gpus_distributed():
    """Test aggregate_metrics_across_gpus in distributed setting"""
    accelerator = Accelerator()

    rank = accelerator.process_index
    metric_tensor = torch.tensor([rank + 1.0], device=accelerator.device)

    result = aggregate_metrics_across_gpus(accelerator, metric_tensor)

    expected_mean = (
        sum(range(1, accelerator.num_processes + 1)) / accelerator.num_processes
    )
    assert (
        abs(result - expected_mean) < 1e-6
    )  # Allow for small floating point differences


def test_aggregate_metrics_across_gpus_with_negative_values():
    """Test aggregate_metrics_across_gpus with negative values"""
    accelerator = Accelerator()

    metric_tensor = torch.tensor([-1.0, -2.0, -3.0], device=accelerator.device)

    result = aggregate_metrics_across_gpus(accelerator, metric_tensor)

    assert result == -2.0  # (-1 + -2 + -3) / 3 = -2.0
    assert isinstance(result, float)


def test_aggregate_metrics_across_gpus_with_zero_values():
    """Test aggregate_metrics_across_gpus with zero values"""
    accelerator = Accelerator()

    metric_tensor = torch.tensor([0.0, 0.0, 0.0], device=accelerator.device)

    result = aggregate_metrics_across_gpus(accelerator, metric_tensor)

    assert result == 0.0
    assert isinstance(result, float)


def test_consolidate_mutations_warning_if_not_llm_algorithm():
    """Test consolidate_mutations"""
    population = [Mock() for _ in range(3)]
    with pytest.warns(UserWarning):
        consolidate_mutations(population)


def test_consolidate_mutations():
    population = [MagicMock(spec=GRPO) for _ in range(3)]
    for agent in population:
        agent.mut = "lr"
        agent.lr = 0.01
        agent.optimizer = Mock()
        agent.optimizer.param_groups = [{"lr": 0.01}]
        agent.cosine_lr_schedule_config = {"warmup_steps": 0, "total_steps": 100}
        agent.accelerator = MagicMock(spec=Accelerator)
        agent.accelerator.is_main_process = True
        agent.accelerator.wait_for_everyone = Mock()
        agent.accelerator.state = MagicMock()
        agent.accelerator.state.deepspeed_plugin = MagicMock(spec=DeepSpeedPlugin)
        agent.accelerator.state.deepspeed_plugin.deepspeed_config = {}
        agent.actor = MagicMock()
    consolidate_mutations(population)
    for agent in population:
        assert agent.mut == "lr"
        assert agent.lr == 0.01
        assert agent.optimizer.param_groups[0]["lr"] == 0.01


def test_tournament_selection_and_mutation_language_model():
    """Test tournament_selection_and_mutation with language model"""
    population = [MagicMock(spec=LLMAlgorithm) for _ in range(3)]
    for agent in population:
        agent.mut = "lr"
        agent.lr = 0.01
        agent.optimizer = Mock()
        agent.optimizer.param_groups = [{"lr": 0.01}]
        agent.accelerator = MagicMock(spec=Accelerator)
        agent.actor = MagicMock()
        agent.actor.save_checkpoint = Mock()
    tournament = MagicMock(spec=TournamentSelection)
    mutation = MagicMock(spec=Mutations)
    mutation.mutation = Mock(return_value=population)
    tournament.select = Mock(return_value=(population[0], population))
    env_name = "CartPole-v1"
    algo = None
    elite_path = None
    save_elite = True
    language_model = True
    accelerator = MagicMock(spec=Accelerator)
    accelerator.is_main_process = True
    accelerator.wait_for_everyone = Mock()

    with (
        patch("agilerl.utils.utils.save_llm_checkpoint") as mock_save_llm_checkpoint,
        patch(
            "agilerl.utils.utils.consolidate_mutations"
        ) as mock_consolidate_mutations,
    ):
        output_pop = tournament_selection_and_mutation(
            population,
            tournament,
            mutation,
            env_name,
            algo,
            elite_path,
            save_elite,
            accelerator,
            language_model,
        )
        mock_save_llm_checkpoint.assert_called_once_with(population[0], elite_path)
        mock_consolidate_mutations.assert_called_once_with(output_pop)

    tournament.select.assert_called_once_with(population)
    mutation.mutation.assert_called_once_with(population)
    accelerator.wait_for_everyone.assert_called()



from dataclasses import asdict

import numpy as np
import pytest
import torch
import torch.nn.functional as F
from gymnasium import spaces

from agilerl.modules import (
    EvolvableCNN,
    EvolvableLSTM,
    EvolvableMLP,
    EvolvableModule,
    EvolvableMultiInput,
    EvolvableSimBa,
)
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.actors import DeterministicActor, StochasticActor
from agilerl.networks.base import EvolvableNetwork
from tests.helper_functions import (
    assert_close_dict,
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    check_equal_params_ind,
)


@pytest.fixture(scope="module")
def head_config():
    return asdict(MlpNetConfig(hidden_size=[64, 64]))


@pytest.mark.parametrize("action_space", ["vector_space"])
@pytest.mark.parametrize(
    "observation_space, encoder_type",
    [
        ("dict_space", "multi_input"),
        ("discrete_space", "mlp"),
        ("vector_space", "mlp"),
        ("image_space", "cnn"),
    ],
)
def test_deterministic_actor_initialization(
    observation_space, action_space, encoder_type, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = DeterministicActor(observation_space, action_space)

    assert network.observation_space == observation_space

    if encoder_type == "multi_input":
        assert isinstance(network.encoder, EvolvableMultiInput)
    elif encoder_type == "mlp":
        assert isinstance(network.encoder, EvolvableMLP)
    elif encoder_type == "cnn":
        assert isinstance(network.encoder, EvolvableCNN)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


def test_deterministic_actor_initialization_recurrent(vector_space):
    observation_space = spaces.Box(low=-1, high=1, shape=(32,))
    network = DeterministicActor(observation_space, vector_space, recurrent=True)

    assert network.observation_space == observation_space
    assert isinstance(network.encoder, EvolvableLSTM)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


def test_deterministic_actor_initialization_simba(vector_space):
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))
    network = DeterministicActor(observation_space, vector_space, simba=True)

    assert network.observation_space == observation_space
    assert isinstance(network.encoder, EvolvableSimBa)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


@pytest.mark.parametrize("action_space", ["vector_space"])
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_deterministic_actor_mutation_methods(
    observation_space, action_space, head_config, dummy_rng, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = DeterministicActor(
        observation_space, action_space, head_config=head_config
    )
    network.rng = dummy_rng

    for method in network.mutation_methods:
        new_network = network.clone()
        getattr(new_network, method)()

        if "." in method:
            net_name = method.split(".")[0]
            mutated_module: EvolvableModule = getattr(new_network, net_name)
            exec_method = new_network.last_mutation_attr.split(".")[-1]

            if isinstance(observation_space, (spaces.Tuple, spaces.Dict)):
                mutated_attr = mutated_module.last_mutation_attr.split(".")[-1]
            else:
                mutated_attr = mutated_module.last_mutation_attr

            assert mutated_attr == exec_method

        if new_network.last_mutation_attr is not None:
            # Check that architecture has changed
            assert_not_equal_state_dict(network.state_dict(), new_network.state_dict())

            # Checks that parameters that are not mutated are the same
            check_equal_params_ind(network, new_network)
        else:
            raise ValueError(
                f"Last mutation attribute is None. Expected {method} to be applied."
            )


@pytest.mark.parametrize("action_space", ["vector_space"])
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_deterministic_actor_forward(
    observation_space: spaces.Space, action_space: spaces.Space, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = DeterministicActor(observation_space, action_space)

    x_np = observation_space.sample()

    if isinstance(observation_space, spaces.Discrete):
        x_np = (
            F.one_hot(torch.tensor(x_np), num_classes=observation_space.n)
            .float()
            .numpy()
        )

    with torch.no_grad():
        out = network(x_np)

    assert isinstance(out, torch.Tensor)
    assert out.shape == torch.Size([1, spaces.flatdim(action_space)])

    if isinstance(observation_space, spaces.Dict):
        x = {key: torch.tensor(value) for key, value in x_np.items()}
    elif isinstance(observation_space, spaces.Tuple):
        x = tuple(torch.tensor(value) for value in x_np)
    else:
        x = torch.tensor(x_np)

    with torch.no_grad():
        out = network(x)

    assert isinstance(out, torch.Tensor)
    assert out.shape == torch.Size([1, spaces.flatdim(action_space)])


@pytest.mark.parametrize("action_space", ["vector_space"])
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_deterministic_actor_clone(
    observation_space: spaces.Space, action_space: spaces.Space, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = DeterministicActor(observation_space, action_space)

    original_net_dict = dict(network.named_parameters())
    clone = network.clone()
    assert isinstance(clone, EvolvableNetwork)

    assert_close_dict(network.init_dict, clone.init_dict)

    assert_state_dicts_equal(clone.state_dict(), network.state_dict())
    for key, param in clone.named_parameters():
        torch.testing.assert_close(param, original_net_dict[key])


def test_deterministic_actor_rescale_action():
    # Test rescaling with different output activations

    # Create a simple action space
    action_low = torch.tensor([-2.0, -1.0], dtype=torch.float64)
    action_high = torch.tensor([2.0, 3.0], dtype=torch.float64)

    # Test with Tanh activation (default for DeterministicActor)
    action = torch.tensor(
        [[-0.5, 0.5]], dtype=torch.float32
    )  # Action from network between -1 and 1
    rescaled = DeterministicActor.rescale_action(
        action, action_low, action_high, "Tanh"
    )

    # Calculate expected value: low + (high - low) * (action - (-1)) / (1 - (-1))
    expected = torch.tensor([[-1.0, 2.0]])  # Should be mapped to middle of range
    torch.testing.assert_close(rescaled, expected)
    assert rescaled.dtype == action.dtype

    # Test with Sigmoid activation
    action = torch.tensor(
        [[0.25, 0.75]], dtype=torch.float32
    )  # Action from network between 0 and 1
    rescaled = DeterministicActor.rescale_action(
        action, action_low, action_high, "Sigmoid"
    )
    expected = torch.tensor(
        [[-1.0, 2.0]]
    )  # Should be mapped to quarter/three-quarters of range
    torch.testing.assert_close(rescaled, expected)
    assert rescaled.dtype == action.dtype

    # Test with no activation (unbounded)
    action = torch.tensor([[1.0, -3.0]], dtype=torch.float32)  # Unbounded action
    rescaled = DeterministicActor.rescale_action(action, action_low, action_high, None)
    # With no activation, action is just passed through as is
    expected = action
    torch.testing.assert_close(rescaled, expected)
    assert rescaled.dtype == action.dtype

    # Test clipping behavior (out-of-bounds values)
    action = torch.tensor(
        [[-2.0, 2.0]], dtype=torch.float32
    )  # Actions outside of tanh range
    rescaled = DeterministicActor.rescale_action(
        action, action_low, action_high, "Tanh"
    )
    # For actions outside the normal range, we still apply the same rescaling formula
    # For -2.0: low + (high - low) * (-2.0 - (-1)) / (1 - (-1)) = low + (high - low) * (-1.0/2.0)
    # For 2.0: low + (high - low) * (2.0 - (-1)) / (1 - (-1)) = low + (high - low) * (3.0/2.0)
    expected_first = action_low + (action_high - action_low) * (
        action[0, 0] - (-1.0)
    ) / (1.0 - (-1.0))
    expected_second = action_low + (action_high - action_low) * (
        action[0, 1] - (-1.0)
    ) / (1.0 - (-1.0))
    expected = torch.tensor(
        [[expected_first[0], expected_second[1]]], dtype=action.dtype
    )
    torch.testing.assert_close(rescaled, expected)
    assert rescaled.dtype == action.dtype


def test_deterministic_actor_forward_rescaling():
    # Test that the forward method rescales actions correctly
    # Create a continuous action space
    action_space = spaces.Box(low=np.array([-2.0, -1.0]), high=np.array([2.0, 3.0]))

    # Test with default Tanh activation
    actor = DeterministicActor(
        observation_space=spaces.Box(low=-1, high=1, shape=(2,)),
        action_space=action_space,
    )

    # Create sample observation
    obs = torch.tensor([[-0.5, 0.5]]).float()

    # Override the head_net to output a known value
    def mock_forward(x):
        return torch.tensor([[-0.5, 0.5]])  # Fixed output for testing

    # Save original function
    original_head_forward = actor.head_net.forward

    try:
        # Mock the head network output
        actor.head_net.forward = mock_forward

        # Get action from forward pass
        with torch.no_grad():
            action = actor(obs)

        # Expected value based on rescaling from [-1,1] to action space
        expected = torch.tensor([[-1.0, 2.0]])  # Middle of the action range
        torch.testing.assert_close(action, expected)

        # Test with explicit Sigmoid activation
        actor = DeterministicActor(
            observation_space=spaces.Box(low=-1, high=1, shape=(2,)),
            action_space=action_space,
            head_config={"output_activation": "Sigmoid", "hidden_size": [32]},
        )

        # Override again for the new actor
        actor.head_net.forward = lambda x: torch.tensor([[0.25, 0.75]])

        with torch.no_grad():
            action = actor(obs)

        expected = torch.tensor(
            [[-1.0, 2.0]]
        )  # Should map to quarter/three-quarters of range
        torch.testing.assert_close(action, expected)

    finally:
        # Restore original function if needed
        if "original_head_forward" in locals():
            actor.head_net.forward = original_head_forward


@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_distribution_mutation_methods(
    dummy_rng, head_config, use_experimental_distribution
):
    observation_space = spaces.Box(low=-1, high=1, shape=(2,))
    action_space = spaces.Box(low=-1, high=1, shape=(2,))

    network = StochasticActor(
        observation_space,
        action_space,
        head_config=head_config,
        use_experimental_distribution=use_experimental_distribution,
    )

    evolvable_dist = network.head_net
    evolvable_dist.rng = dummy_rng

    for method in evolvable_dist.mutation_methods:
        new_dist = evolvable_dist.clone()
        getattr(new_dist, method)()

        if "." in method and new_dist.last_mutation_attr is not None:
            net_name = method.split(".")[0]
            mutated_module: EvolvableModule = getattr(new_dist, net_name)

            exec_method = new_dist.last_mutation_attr.split(".")[-1]
            mutated_attr = mutated_module.last_mutation_attr.split(".")[-1]

            assert mutated_attr == exec_method

        if new_dist.last_mutation_attr is not None:
            # Check that architecture has changed
            assert_not_equal_state_dict(
                evolvable_dist.state_dict(), new_dist.state_dict()
            )

            # Checks that parameters that are not mutated are the same
            check_equal_params_ind(evolvable_dist, new_dist)
        else:
            raise ValueError(
                f"Last mutation attribute is None. Expected {method} to be applied."
            )


@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize(
    "observation_space, encoder_type",
    [
        ("dict_space", "multi_input"),
        ("discrete_space", "mlp"),
        ("vector_space", "mlp"),
        ("image_space", "cnn"),
    ],
)
@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_stochastic_actor_initialization(
    observation_space,
    action_space,
    encoder_type,
    use_experimental_distribution,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = StochasticActor(
        observation_space,
        action_space,
        use_experimental_distribution=use_experimental_distribution,
    )

    assert network.observation_space == observation_space

    if encoder_type == "multi_input":
        assert isinstance(network.encoder, EvolvableMultiInput)
    elif encoder_type == "mlp":
        assert isinstance(network.encoder, EvolvableMLP)
    elif encoder_type == "cnn":
        assert isinstance(network.encoder, EvolvableCNN)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


def test_stochastic_actor_initialization_recurrent(vector_space):
    observation_space = spaces.Box(low=-1, high=1, shape=(32, 8))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        network = StochasticActor(
            observation_space,
            vector_space,
            recurrent=True,
            use_experimental_distribution=use_experimental_distribution,
        )

        assert network.observation_space == observation_space
        assert isinstance(network.encoder, EvolvableLSTM)

        evolvable_modules = network.modules()
        assert "encoder" in evolvable_modules
        assert "head_net" in evolvable_modules


def test_stochastic_actor_initialization_simba(vector_space):
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        network = StochasticActor(
            observation_space,
            vector_space,
            simba=True,
            use_experimental_distribution=use_experimental_distribution,
        )

        assert network.observation_space == observation_space
        assert isinstance(network.encoder, EvolvableSimBa)

        evolvable_modules = network.modules()
        assert "encoder" in evolvable_modules
        assert "head_net" in evolvable_modules


@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_stochastic_actor_mutation_methods(
    observation_space,
    action_space,
    head_config,
    dummy_rng,
    use_experimental_distribution,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = StochasticActor(
        observation_space,
        action_space,
        head_config=head_config,
        use_experimental_distribution=use_experimental_distribution,
    )
    network.rng = dummy_rng

    for method in network.mutation_methods:
        new_network = network.clone()
        getattr(new_network, method)()

        if "." in method and new_network.last_mutation_attr is not None:
            net_name = method.split(".")[0]
            mutated_module: EvolvableModule = getattr(new_network, net_name)

            exec_method = new_network.last_mutation_attr.split(".")[-1]

            if isinstance(observation_space, (spaces.Tuple, spaces.Dict)):
                mutated_attr = mutated_module.last_mutation_attr.split(".")[-1]
            else:
                mutated_attr = mutated_module.last_mutation_attr

            assert mutated_attr == exec_method

        if new_network.last_mutation_attr is not None:
            # Check that architecture has changed
            assert_not_equal_state_dict(network.state_dict(), new_network.state_dict())

            # Checks that parameters that are not mutated are the same
            check_equal_params_ind(network, new_network)
        else:
            raise ValueError(
                f"Last mutation attribute is None. Expected {method} to be applied."
            )


@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_stochastic_actor_forward(
    observation_space: spaces.Space,
    action_space: spaces.Space,
    use_experimental_distribution: bool,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    # For continuous action spaces, we need to set squash_output=True to ensure actions are scaled
    squash_output = isinstance(action_space, spaces.Box)

    network = StochasticActor(
        observation_space,
        action_space,
        squash_output=squash_output,
        use_experimental_distribution=use_experimental_distribution,
    )

    x_np = observation_space.sample()

    if isinstance(observation_space, spaces.Discrete):
        x_np = (
            F.one_hot(torch.tensor(x_np), num_classes=observation_space.n)
            .float()
            .numpy()
        )

    with torch.no_grad():
        action, log_prob, entropy = network(x_np)

    # Check action shape and values
    assert isinstance(action, torch.Tensor)

    # For continuous action spaces
    if isinstance(action_space, spaces.Box):
        # Check that actions are within bounds
        action_low = torch.tensor(action_space.low, device=action.device)
        action_high = torch.tensor(action_space.high, device=action.device)
        assert torch.all(action >= action_low)
        assert torch.all(action <= action_high)

    # For discrete action spaces
    elif isinstance(action_space, spaces.Discrete):
        assert action.shape == torch.Size([1])  # One scalar action
        assert torch.all(action >= 0)
        assert torch.all(action < action_space.n)

    # For multi-discrete action spaces
    elif isinstance(action_space, spaces.MultiDiscrete):
        assert action.shape == torch.Size([1, len(action_space.nvec)])
        for i, n in enumerate(action_space.nvec):
            assert torch.all(action[:, i] >= 0)
            assert torch.all(action[:, i] < n)

    # For multi-binary action spaces
    elif isinstance(action_space, spaces.MultiBinary):
        assert action.shape == torch.Size([1, action_space.n])
        assert torch.all((action == 0) | (action == 1))

    # Check log_prob
    assert isinstance(log_prob, torch.Tensor)

    # Check entropy - it might be None for some distribution types
    if entropy is not None:
        assert isinstance(entropy, torch.Tensor)

    # Test with tensor inputs
    if isinstance(observation_space, spaces.Dict):
        x = {key: torch.tensor(value).clone().detach() for key, value in x_np.items()}
    elif isinstance(observation_space, spaces.Tuple):
        x = tuple(torch.tensor(value).clone().detach() for value in x_np)
    else:
        x = torch.tensor(x_np).clone().detach()

    with torch.no_grad():
        action, log_prob, entropy = network(x)

    assert isinstance(action, torch.Tensor)
    assert isinstance(log_prob, torch.Tensor)
    # Entropy might be None
    if entropy is not None:
        assert isinstance(entropy, torch.Tensor)


@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_stochastic_actor_clone(
    observation_space: spaces.Space,
    action_space: spaces.Space,
    use_experimental_distribution: bool,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    network = StochasticActor(
        observation_space,
        action_space,
        use_experimental_distribution=use_experimental_distribution,
    )

    original_net_dict = dict(network.named_parameters())
    clone = network.clone()
    assert isinstance(clone, EvolvableNetwork)

    assert_close_dict(network.init_dict, clone.init_dict)

    assert_state_dicts_equal(clone.state_dict(), network.state_dict())
    for key, param in clone.named_parameters():
        torch.testing.assert_close(param, original_net_dict[key])


def test_stochastic_actor_scaling():
    # Test the scale_action method with different squash_output values

    # Create a continuous action space
    action_space = spaces.Box(low=np.array([-2.0, -1.0]), high=np.array([2.0, 3.0]))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        # Test with squash_output=True
        actor = StochasticActor(
            observation_space=spaces.Box(low=-1, high=1, shape=(2,)),
            action_space=action_space,
            squash_output=True,
            use_experimental_distribution=use_experimental_distribution,
        )

        # Test rescaling from [-1, 1] to [low, high]
        # Use action with same dimension as action space (2)
        action = torch.tensor([[-1.0, 0.0], [1.0, -1.0]])  # Batch of actions
        scaled = actor.scale_action(action)

        # Expected: For -1 -> low, for 0 -> middle, for 1 -> high
        expected = torch.tensor([[-2.0, 1.0], [2.0, -1.0]])
        torch.testing.assert_close(scaled, expected)

        # Test with forward method when squash_output=True
        obs = torch.tensor([[-0.5, 0.5]]).float()

        # Mock the distribution to return a known action
        original_forward = actor.head_net.forward

        def mock_forward(x, mask=None):
            return torch.tensor([[-0.5, 0.5]]), torch.tensor([0.0]), torch.tensor([1.0])

        try:
            actor.head_net.forward = mock_forward

            # Get action from forward pass
            with torch.no_grad():
                action, _, _ = actor(obs)

            # Expected value based on rescaling from [-1,1] to action space
            # For -0.5: low + (0.5 * (-0.5 + 1) * (high - low)) = low + 0.25 * (high - low)
            # For 0.5: low + (0.5 * (0.5 + 1) * (high - low)) = low + 0.75 * (high - low)
            expected = torch.tensor(
                [[-1.0, 2.0]]
            )  # Should be between low and middle, middle and high
            torch.testing.assert_close(action, expected)

        finally:
            # Restore original function
            actor.head_net.forward = original_forward

        # Test with discrete action space (no scaling needed)
        discrete_space = spaces.Discrete(3)
        actor = StochasticActor(
            observation_space=spaces.Box(low=-1, high=1, shape=(2,)),
            action_space=discrete_space,
            use_experimental_distribution=use_experimental_distribution,
        )

        # For discrete spaces, no scaling is applied - verify by testing forward
        def mock_discrete_forward(x, mask=None):
            return torch.tensor([1]), torch.tensor([0.0]), torch.tensor([1.0])

        try:
            actor.head_net.forward = mock_discrete_forward

            with torch.no_grad():
                action, _, _ = actor(obs)

            # Action should remain unchanged
            expected = torch.tensor([1])
            torch.testing.assert_close(action, expected)
        finally:
            # Restore original function
            actor.head_net.forward = mock_discrete_forward


@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize("use_experimental_distribution", [False, True])
def test_stochastic_actor_distribution_methods(
    action_space: spaces.Space, use_experimental_distribution: bool, request
):
    """Test StochasticActor's distribution-related methods with different action spaces."""
    action_space = request.getfixturevalue(action_space)
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))
    network = StochasticActor(
        observation_space,
        action_space,
        use_experimental_distribution=use_experimental_distribution,
    )

    # Get a sample observation
    obs = torch.tensor(observation_space.sample()).float().unsqueeze(0)

    # Get action, log_prob, and entropy from forward pass
    with torch.no_grad():
        action, log_prob, entropy = network(obs)

    # Test action_entropy matches the entropy from forward
    manual_entropy = network.action_entropy()
    torch.testing.assert_close(entropy, manual_entropy)

    # Specific tests for each action space type
    if isinstance(action_space, spaces.Box):
        # For continuous spaces, log_prob should be scalar
        assert log_prob.shape == torch.Size([1])
        # Entropy should be positive for continuous distributions
        assert torch.all(entropy > 0)

    elif isinstance(action_space, spaces.Discrete):
        # For discrete spaces, log_prob is scalar
        assert log_prob.shape == torch.Size([1])
        # Entropy should be positive and at most log(n)
        assert torch.all(entropy > 0)
        assert torch.all(
            entropy <= torch.log(torch.tensor(action_space.n, dtype=torch.float))
        )

    elif isinstance(action_space, spaces.MultiDiscrete):
        # For multi-discrete spaces, log_prob is scalar (sum of log probs)
        assert log_prob.shape == torch.Size([1])
        # Entropy should be positive
        assert torch.all(entropy > 0)

    elif isinstance(action_space, spaces.MultiBinary):
        # For multi-binary spaces, log_prob is scalar
        assert log_prob.shape == torch.Size([1])
        # Entropy should be positive
        assert torch.all(entropy > 0)

    # Test action_log_prob gives value for the chosen action
    # Note: This may not give exactly the same value as from forward
    # since the distribution is resampled, but we can check it has the right shape
    manual_log_prob = network.action_log_prob(action)
    assert manual_log_prob.shape == log_prob.shape


def test_stochastic_actor_action_masking_discrete():
    """Test action masking in StochasticActor with Discrete action space."""
    # Create discrete action space (4 actions)
    action_space = spaces.Discrete(4)
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        # Create actor
        actor = StochasticActor(
            observation_space,
            action_space,
            use_experimental_distribution=use_experimental_distribution,
        )

        # Sample observation
        obs = torch.tensor(observation_space.sample()).float().unsqueeze(0)

        # Create action mask that only allows actions 1 and 3
        # True = valid action, False = masked action
        action_mask = torch.tensor([[False, True, False, True]])

        # Run forward pass with mask
        with torch.no_grad():
            action, log_prob, entropy = actor(obs, action_mask)

        # Verify that the selected action is either 1 or 3 (unmasked)
        assert action.item() in [1, 3]

        # Run many times to ensure consistent masking
        valid_actions = []
        for _ in range(10):
            with torch.no_grad():
                action, _, _ = actor(obs, action_mask)
            valid_actions.append(action.item())

        # Verify only unmasked actions were selected
        assert all(a in [1, 3] for a in valid_actions)

        # The log probabilities of masked actions should be -inf
        # This is hard to test directly without accessing the distribution
        # but we can verify entropy is smaller with masks
        with torch.no_grad():
            _, _, unmasked_entropy = actor(obs, None)  # No mask
            _, _, masked_entropy = actor(obs, action_mask)  # With mask

        # Entropy should be lower with mask (fewer options)
        assert masked_entropy < unmasked_entropy


def test_stochastic_actor_action_masking_multidiscrete():
    """Test action masking in StochasticActor with MultiDiscrete action space."""
    # Create multi-discrete action space
    action_space = spaces.MultiDiscrete([3, 2, 4])
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        # Create actor
        actor = StochasticActor(
            observation_space,
            action_space,
            use_experimental_distribution=use_experimental_distribution,
        )

        # Sample observation
        obs = torch.tensor(observation_space.sample()).float().unsqueeze(0)

        # Create action masks for each dimension
        # Dimension 1: allow only action 0
        # Dimension 2: allow only action 1
        # Dimension 3: allow actions 1 and 3
        # Format as a flat tensor that will be split by the distribution
        action_mask = torch.tensor(
            [
                [
                    True,
                    False,
                    False,  # 1st dimension (3 options)
                    False,
                    True,  # 2nd dimension (2 options)
                    False,
                    True,
                    False,
                    True,
                ]  # 3rd dimension (4 options)
            ]
        )

        # Run forward pass with mask multiple times
        valid_actions = []
        for _ in range(10):
            with torch.no_grad():
                action, _, _ = actor(obs, action_mask)
            valid_actions.append(action.squeeze().tolist())

        # Verify all actions respect the mask
        for action in valid_actions:
            assert action[0] == 0  # 1st dimension must be 0
            assert action[1] == 1  # 2nd dimension must be 1
            assert action[2] in [1, 3]  # 3rd dimension must be 1 or 3


def test_stochastic_actor_action_masking_multibinary():
    """Test action masking in StochasticActor with MultiBinary action space."""
    # Create multi-binary action space
    action_space = spaces.MultiBinary(4)
    observation_space = spaces.Box(low=-1, high=1, shape=(8,))

    # Test with both default and experimental distribution
    for use_experimental_distribution in [False, True]:
        # Create actor
        actor = StochasticActor(
            observation_space,
            action_space,
            use_experimental_distribution=use_experimental_distribution,
        )

        # Sample observation
        obs = torch.tensor(observation_space.sample()).float().unsqueeze(0)

        # For MultiBinary, the mask should have the same shape as the action space
        # True means this action is allowed, False means it's masked (forbidden)
        # We'll mask the first and last bits to only allow action=0
        action_mask = torch.tensor([[False, True, True, False]])

        # Run forward pass with mask multiple times
        valid_actions = []
        for _ in range(10):
            with torch.no_grad():
                action, _, _ = actor(obs, action_mask)
            valid_actions.append(action.squeeze().tolist())

        # Verify all actions respect the mask
        for action in valid_actions:
            # The first and last bits should be 0 because we masked them to prevent 1
            assert action[0] == 0
            assert action[3] == 0
            # Positions 1 and 2 can be either 0 or 1



from enum import Enum
from numbers import Number
from typing import (
    Any,
    Callable,
    ClassVar,
    Protocol,
    TypedDict,
    TypeVar,
    Union,
)

import gymnasium as gym
import numpy as np
import torch
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from pettingzoo import ParallelEnv
from tensordict import TensorDict
from torch._dynamo import OptimizedModule
from torch.nn import Module
from torch.optim import Optimizer

from agilerl.protocols import (
    EvolvableAlgorithm,
    EvolvableModule,
    EvolvableNetwork,
    ModuleDict,
)

# Type variable for module types - bound to Module to ensure all types inherit from it
T = TypeVar("T", bound=Union[Module, OptimizedModule])


class IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Any]]


# TODO ideally adjust the ReasoningGym to match the PreferencePrompts type (e.g the lists/btaching happens for the values of the
# keys, not just returning a list of 'ReasoningPrompts')
class ReasoningPrompts(TypedDict):
    input_ids: torch.Tensor
    attention_mask: torch.Tensor
    text: str | None


class PreferencePrompts(TypedDict):
    prompt: list[str]
    prompt_lengths: list[int]
    chosen: list[str]
    rejected: list[str]
    chosen_input_ids: torch.Tensor
    chosen_attention_mask: torch.Tensor
    rejected_input_ids: torch.Tensor
    rejected_attention_mask: torch.Tensor


class MultiAgentSetup(Enum):
    """Enum to specify the type of multi-agent setup."""

    HOMOGENEOUS = "homogeneous"  # all agents have the same network architecture
    MIXED = "mixed"  # contains a mix of different network architectures
    HETEROGENEOUS = "heterogeneous"  # all agents have different network architectures


class ModuleType(Enum):
    """Enum to specify the type of module."""

    MLP = "mlp"
    CNN = "cnn"
    RNN = "rnn"
    MULTI_INPUT = "multi_input"


SupportedObsSpaces = Union[
    spaces.Box,
    spaces.Discrete,
    spaces.MultiDiscrete,
    spaces.Dict,
    spaces.Tuple,
    spaces.MultiBinary,
]

SupportedActionSpaces = Union[
    spaces.Discrete,
    spaces.MultiDiscrete,
    spaces.MultiBinary,
    spaces.Box,
]

ArrayOrTensor = Union[np.ndarray, torch.Tensor]
StandardTensorDict = dict[str, torch.Tensor]
TensorTuple = tuple[torch.Tensor, ...]
ArrayDict = dict[str, np.ndarray]
ArrayTuple = tuple[np.ndarray, ...]
NetConfigType = dict[str, Union[dict[str, Any], Any]]
KernelSizeType = Union[int, tuple[int, ...]]
GymSpaceType = Union[SupportedObsSpaces, list[SupportedObsSpaces]]
GymEnvType = Union[str, gym.Env, gym.vector.VectorEnv, gym.vector.AsyncVectorEnv]
PzEnvType = Union[str, ParallelEnv]
LLMObsType = list[ReasoningPrompts]

NumpyObsType = Union[np.ndarray, ArrayDict, ArrayTuple]
TorchObsType = Union[torch.Tensor, TensorDict, TensorTuple, StandardTensorDict]
ObservationType = Union[NumpyObsType, TorchObsType, Number, LLMObsType]
MultiAgentObservationType = dict[str, ObservationType]
ActionType = Union[int, float, np.ndarray, torch.Tensor]
InfosDict = dict[str, dict[str, Any]]
MaybeObsList = Union[list[ObservationType], ObservationType]
ExperiencesType = Union[dict[str, ObservationType], tuple[ObservationType, ...]]
ActionReturnType = Union[tuple[Union[ActionType, Any], ...], ActionType, Any]
GymStepReturn = tuple[NumpyObsType, ActionType, float, MaybeObsList, InfosDict]
PzStepReturn = tuple[
    dict[str, NumpyObsType], ArrayDict, ArrayDict, ArrayDict, dict[str, Any]
]

SingleAgentModule = Union[T, EvolvableModule, OptimizedModule, EvolvableNetwork]
MultiAgentModule = ModuleDict[SingleAgentModule[T]]
NetworkType = Union[SingleAgentModule[T], MultiAgentModule[T]]
EvolvableNetworkType = Union[EvolvableModule, ModuleDict[EvolvableModule]]
DeviceType = Union[str, torch.device]
OptimizerType = Union[Optimizer, AcceleratedOptimizer]
ConfigType = Union[IsDataclass, NetConfigType]
StateDict = Union[dict[str, Any], dict[str, dict[str, Any]]]

SingleAgentMutReturnType = dict[str, Any]
MultiAgentMutReturnType = dict[str, dict[str, Any]]
MutationReturnType = Union[SingleAgentMutReturnType, MultiAgentMutReturnType]
PopulationType = list[EvolvableAlgorithm]
MutationMethod = Callable[[EvolvableAlgorithm], EvolvableAlgorithm]
ConfigType = Union[IsDataclass, dict[str, Any]]
StateDict = Union[dict[str, Any], list[dict[str, Any]]]


class BatchDimension:
    def __repr__(self):
        return "BatchDimension"


class BPTTSequenceType(Enum):
    """Enum for BPTT sequence generation methods. It specifies the strategy used when generating sequences for BPTT training.

    CHUNKED is the default method which uses the least amount of memory while keeping all sampled trajectories available in the buffer for sequencing.
        The number of sequences generated is then:  (num_steps / max_seq_len) * num_envs
    MAXIMUM generates all possible overlapping sequences, which is the most memory-intensive option.
        The number of sequences generated is then:  (num_steps - max_seq_len + 1) * num_envs
    FIFTY_PERCENT_OVERLAP generates sequences with 50% overlap, which is a compromise between the two.
        The number of sequences generated is then:  (num_steps / max_seq_len * 2) * num_envs
    """

    CHUNKED = "chunked"  # Generate sequences by non-overlapping chunks
    MAXIMUM = "maximum"  # Generate all possible overlapping sequences
    FIFTY_PERCENT_OVERLAP = (
        "fifty_percent_overlap"  # Generate sequences with 50% overlap
    )



import numpy as np
from accelerate.utils import broadcast_object_list

from agilerl.algorithms.core.base import EvolvableAlgorithm

PopulationType = list[EvolvableAlgorithm]


class TournamentSelection:
    """The tournament selection class. Calling :func:`TournamentSelection.select() <agilerl.hpo.tournament.TournamentSelection.select>`
    on a population of agents will return a cloned population containing the best performing agent as well as the new generation of agents
    based on their fitness scores.

    :param tournament_size: Tournament selection size
    :type tournament_size: int
    :param elitism: Elitism in tournament selection
    :type elitism: bool
    :param population_size: Number of agents in population
    :type population_size: int
    :param eval_loop: Number of most recent fitness scores to use in evaluation
    :type eval_loop: int
    """

    def __init__(
        self,
        tournament_size: int,
        elitism: bool,
        population_size: int,
        eval_loop: int,
    ) -> None:
        assert tournament_size > 0, "Tournament size must be greater than zero."
        assert isinstance(elitism, bool), "Elitism must be boolean value True or False."
        assert population_size > 0, "Population size must be greater than zero."
        assert eval_loop > 0, "Evo step must be greater than zero."
        self.tournament_size = tournament_size
        self.elitism = elitism
        self.population_size = population_size
        self.eval_loop = eval_loop
        self.language_model = None

    def _tournament(self, fitness_values: list[float]) -> int:
        """
        Perform tournament selection given a list of fitness values.

        :param fitness_values: List of fitness values
        :type fitness_values: list[float]
        :return: Index of the selected winner
        :rtype: int
        """
        selection = np.random.randint(0, len(fitness_values), size=self.tournament_size)
        selection_values = [fitness_values[i] for i in selection]
        winner = selection[np.argmax(selection_values)]
        return winner

    def _elitism(
        self, population: PopulationType
    ) -> tuple[EvolvableAlgorithm, np.ndarray, int]:
        """
        Perform elitism selection given a population of agents.

        :param population: Population of agents
        :type population: PopulationType
        :return: Elite member of population, rank array, and max id
        :rtype: tuple[EvolvableAlgorithm, np.ndarray, int]
        """
        last_fitness = [np.mean(indi.fitness[-self.eval_loop :]) for indi in population]
        rank = np.argsort(last_fitness).argsort()
        max_id = max([ind.index for ind in population])
        model = population[int(np.argsort(rank)[-1])]
        elite = model.clone() if not self.language_model else model.index
        return elite, rank, max_id

    def select(
        self, population: PopulationType
    ) -> tuple[EvolvableAlgorithm, PopulationType]:
        """
        Select the best agent and new population of agents following tournament selection.

        :param population: Population of agents
        :type population: PopulationType
        :return: Elite agent and new population
        :rtype: tuple[EvolvableAlgorithm, PopulationType]
        """
        if self.language_model is None:
            self.language_model = population[0].algo == "GRPO"

        return (
            self._select_llm_agents(population)
            if self.language_model
            else self._select_standard_agents(population)
        )

    def _select_standard_agents(
        self, population: PopulationType
    ) -> tuple[EvolvableAlgorithm, PopulationType]:
        """
        Returns best agent and new population of agents following tournament selection. Used for
        a population of :class:`EvolvableAlgorithm <agilerl.algorithms.core.RLAlgorithm>` or
        :class:`MultiAgentRLAlgorithm <agilerl.algorithms.core.MultiAgentRLAlgorithm>` agents.

        :param population: Population of agents
        :type population: PopulationType
        :return: Elite agent and new population
        :rtype: tuple[EvolvableAlgorithm, PopulationType]
        """
        elite, rank, max_id = self._elitism(population)
        new_population = []
        if self.elitism:  # keep top agent in population
            new_population.append(elite.clone(wrap=False))
            selection_size = self.population_size - 1
        else:
            selection_size = self.population_size

        # Select parents of next gen using tournament selection
        for idx in range(selection_size):
            max_id += 1
            actor_parent = population[self._tournament(rank)]
            new_individual = actor_parent.clone(max_id, wrap=False)
            new_population.append(new_individual)

        return elite, new_population

    def _select_llm_agents(
        self, population: PopulationType
    ) -> tuple[EvolvableAlgorithm, PopulationType]:
        """
        Returns best agent and new population of agents following tournament selection. Used for
        a population of :class:`LLMAlgorithm <agilerl.algorithms.core.LLMAlgorithm>` agents.

        :param population: Population of agents
        :type population: PopulationType
        :return: Elite agent and new population
        :rtype: tuple[EvolvableAlgorithm, PopulationType]
        """
        accelerator = population[0].accelerator
        new_population_idxs = []
        old_population_idxs = [ind.index for ind in population]
        unwanted_agents = {}

        if accelerator is None or (
            accelerator is not None and accelerator.is_main_process
        ):
            elite_idx, rank, max_id = self._elitism(population)
            if self.elitism:  # keep top agent in population
                new_population_idxs.append((elite_idx, elite_idx, True))
                selection_size = self.population_size - 1
            else:
                elite = population[old_population_idxs.index(elite_idx)]
                selection_size = self.population_size
            # select parents of next gen using tournament selection
            for _ in range(selection_size):
                max_id += 1
                actor_parent_idx = old_population_idxs[self._tournament(rank)]
                new_population_idxs.append(
                    (actor_parent_idx, max_id, False)
                )  # (old_idx_to_clone, new_labelled_idx, is_elite)

            # Isolate any agents that are not in the new population to be deleted
            unwanted_agents = set(old_population_idxs) - {
                idx for idx, *_ in new_population_idxs
            }

        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.num_processes > 1:
                new_population_idxs, old_population_idxs, unwanted_agents = (
                    broadcast_object_list(
                        [new_population_idxs, old_population_idxs, unwanted_agents],
                        from_process=0,
                    )
                )

        # Delete any unwanted agents from memory
        for agent_idx in old_population_idxs:
            if agent_idx in unwanted_agents:
                agent_ref = population[old_population_idxs.index(agent_idx)]
                if agent_ref.accelerator is not None:
                    agent_ref.accelerator.wait_for_everyone()
                agent_ref.clean_up()
                if agent_ref.accelerator is not None:
                    agent_ref.accelerator.wait_for_everyone()
                agent_ref = None

        new_population = []
        index_tracker = {}
        for idx_to_clone, new_idx, is_elite in new_population_idxs:
            if (
                agent_ref := population[old_population_idxs.index(idx_to_clone)]
            ) is not None:
                if agent_ref.accelerator is not None:
                    agent_ref.accelerator.wait_for_everyone()
                actor_parent = agent_ref.clone(new_idx, wrap=False)
                if agent_ref.accelerator is not None:
                    agent_ref.accelerator.wait_for_everyone()
                agent_ref.clean_up()
                if agent_ref.accelerator is not None:
                    agent_ref.accelerator.wait_for_everyone()
                agent_ref = population[old_population_idxs.index(idx_to_clone)] = None
                index_tracker[idx_to_clone] = actor_parent
            else:
                actor_parent = index_tracker[idx_to_clone].clone(new_idx, wrap=False)
            if is_elite:
                elite = actor_parent
            new_population.append(actor_parent)
        return elite, new_population



import numpy as np
import torch
from pettingzoo.mpe import simple_speaker_listener_v4
from tqdm import trange

from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.utils import create_population, observation_space_channels_to_first
from agilerl.vector.pz_async_vec_env import AsyncPettingZooVecEnv

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


if __name__ == "__main__":
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("===== AgileRL Online Multi-Agent Demo =====")

    # Define the network configuration
    NET_CONFIG = {
        "encoder_config": {
            "hidden_size": [32, 32],  # Actor hidden size
        }
    }

    # Define the initial hyperparameters
    INIT_HP = {
        # Swap image channels dimension from last to first [H, W, C] -> [C, H, W]
        "CHANNELS_LAST": False,
        "BATCH_SIZE": 32,  # Batch size
        "O_U_NOISE": True,  # Ornstein Uhlenbeck action noise
        "EXPL_NOISE": 0.1,  # Action noise scale
        "MEAN_NOISE": 0.0,  # Mean action noise
        "THETA": 0.15,  # Rate of mean reversion in OU noise
        "DT": 0.01,  # Timestep for OU noise
        "LR_ACTOR": 0.001,  # Actor learning rate
        "LR_CRITIC": 0.001,  # Critic learning rate
        "GAMMA": 0.95,  # Discount factor
        "MEMORY_SIZE": 100000,  # Max memory buffer size
        "LEARN_STEP": 100,  # Learning frequency
        "TAU": 0.01,  # For soft update of target parameters
        "POLICY_FREQ": 2,  # Policy frequnecy
        "POP_SIZE": 4,  # Population size
    }

    num_envs = 8
    # Define the simple speaker listener environment as a parallel environment
    env = AsyncPettingZooVecEnv(
        [
            lambda: simple_speaker_listener_v4.parallel_env(continuous_actions=True)
            for _ in range(num_envs)
        ]
    )
    env.reset()

    # Configure the multi-agent algo input arguments
    observation_spaces = [env.single_observation_space(agent) for agent in env.agents]
    action_spaces = [env.single_action_space(agent) for agent in env.agents]

    if INIT_HP["CHANNELS_LAST"]:
        observation_spaces = [
            observation_space_channels_to_first(space) for space in observation_spaces
        ]

    # Append number of agents and agent IDs to the initial hyperparameter dictionary
    INIT_HP["N_AGENTS"] = env.num_agents
    INIT_HP["AGENT_IDS"] = env.agents

    # Create a population ready for evolutionary hyper-parameter optimisation
    pop = create_population(
        "MADDPG",
        observation_spaces,
        action_spaces,
        NET_CONFIG,
        INIT_HP,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=num_envs,
        device=device,
    )

    # Configure the multi-agent replay buffer
    field_names = ["state", "action", "reward", "next_state", "done"]
    memory = MultiAgentReplayBuffer(
        INIT_HP["MEMORY_SIZE"],
        field_names=field_names,
        agent_ids=INIT_HP["AGENT_IDS"],
        device=device,
    )

    # Instantiate a tournament selection object (used for HPO)
    tournament = TournamentSelection(
        tournament_size=2,  # Tournament selection size
        elitism=True,  # Elitism in tournament selection
        population_size=INIT_HP["POP_SIZE"],  # Population size
        eval_loop=1,  # Evaluate using last N fitness scores
    )

    # Instantiate a mutations object (used for HPO)
    mutations = Mutations(
        algo="MADDPG",
        no_mutation=0.2,  # Probability of no mutation
        architecture=0.2,  # Probability of architecture mutation
        new_layer_prob=0.2,  # Probability of new layer mutation
        parameters=0.2,  # Probability of parameter mutation
        activation=0,  # Probability of activation function mutation
        rl_hp=0.2,  # Probability of RL hyperparameter mutation
        rl_hp_selection=[
            "lr",
            "learn_step",
            "batch_size",
        ],  # RL hyperparams selected for mutation
        mutation_sd=0.1,  # Mutation strength
        agent_ids=INIT_HP["AGENT_IDS"],
        rand_seed=1,
        device=device,
    )

    # Define training loop parameters
    max_steps = 1000000  # Max steps
    learning_delay = 0  # Steps before starting learning

    evo_steps = 10000  # Evolution frequency
    eval_steps = None  # Evaluation steps per episode - go until done
    eval_loop = 1  # Number of evaluation episodes

    total_steps = 0

    # TRAINING LOOP
    print("Training...")
    pbar = trange(max_steps, unit="step")
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        pop_episode_scores = []
        for agent in pop:  # Loop through population
            state, info = env.reset()  # Reset environment at start of episode
            scores = np.zeros(num_envs)
            completed_episode_scores = []
            steps = 0
            if INIT_HP["CHANNELS_LAST"]:
                state = {
                    agent_id: np.moveaxis(s, [-1], [-3])
                    for agent_id, s in state.items()
                }

            for idx_step in range(evo_steps // num_envs):

                # Get next action from agent
                action, raw_action = agent.get_action(
                    states=state, training=True, infos=info
                )

                # Act in environment
                next_state, reward, termination, truncation, info = env.step(action)

                scores += np.sum(np.array(list(reward.values())).transpose(), axis=-1)
                total_steps += num_envs
                steps += num_envs

                # Image processing if necessary for the environment
                if INIT_HP["CHANNELS_LAST"]:
                    next_state = {
                        agent_id: np.moveaxis(ns, [-1], [-3])
                        for agent_id, ns in next_state.items()
                    }

                # Save experiences to replay buffer
                memory.save_to_memory(
                    state,
                    raw_action,
                    reward,
                    next_state,
                    termination,
                    is_vectorised=True,
                )

                # Learn according to learning frequency
                # Handle learn steps > num_envs
                if agent.learn_step > num_envs:
                    learn_step = agent.learn_step // num_envs
                    if (
                        idx_step % learn_step == 0
                        and len(memory) >= agent.batch_size
                        and memory.counter > learning_delay
                    ):
                        # Sample replay buffer
                        experiences = memory.sample(agent.batch_size)
                        # Learn according to agent's RL algorithm
                        agent.learn(experiences)
                # Handle num_envs > learn step; learn multiple times per step in env
                elif (
                    len(memory) >= agent.batch_size and memory.counter > learning_delay
                ):
                    for _ in range(num_envs // agent.learn_step):
                        # Sample replay buffer
                        experiences = memory.sample(agent.batch_size)
                        # Learn according to agent's RL algorithm
                        agent.learn(experiences)

                state = next_state

                # Calculate scores and reset noise for finished episodes
                reset_noise_indices = []
                term_array = np.array(list(termination.values())).transpose()
                trunc_array = np.array(list(truncation.values())).transpose()
                for idx, (d, t) in enumerate(zip(term_array, trunc_array)):
                    if np.any(d) or np.any(t):
                        completed_episode_scores.append(scores[idx])
                        agent.scores.append(scores[idx])
                        scores[idx] = 0
                        reset_noise_indices.append(idx)
                agent.reset_action_noise(reset_noise_indices)

            pbar.update(evo_steps // len(pop))

            agent.steps[-1] += steps
            pop_episode_scores.append(completed_episode_scores)

        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                swap_channels=INIT_HP["CHANNELS_LAST"],
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]
        mean_scores = [
            (
                np.mean(episode_scores)
                if len(episode_scores) > 0
                else "0 completed episodes"
            )
            for episode_scores in pop_episode_scores
        ]

        print(f"--- Global steps {total_steps} ---")
        print(f"Steps {[agent.steps[-1] for agent in pop]}")
        print(f"Scores: {mean_scores}")
        print(f'Fitnesses: {["%.2f"%fitness for fitness in fitnesses]}')
        print(
            f'5 fitness avgs: {["%.2f"%np.mean(agent.fitness[-5:]) for agent in pop]}'
        )

        # Tournament selection and population mutation
        elite, pop = tournament.select(pop)
        pop = mutations.mutation(pop)

        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

    pbar.close()
    env.close()
